from pyVHR.realtime.params import Params
from pyVHR.realtime.VHRroutine import *
import PySimpleGUI as sg
import threading
import ast
import pickle


def GUI_MENU():
    col1 = [[sg.Text("Set Parameters:")],
            [sg.Text("Video file path, or device number:")],
            [sg.In("", key="-VideoFileName-"),
             sg.FileBrowse(target=("-VideoFileName-"))],
            [sg.Text("Use CUDA:"),
             sg.Radio("True", "RADIO1", default=True, key="-cudaTrue-"),
             sg.Radio("False", "RADIO1", default=False, key="-cudaFalse-")],
            [sg.Text("Window Size:"),
             sg.In(str(Params.winSize), key="-winSize-", size=(5, 1))],
            [sg.Text("Stride:"),
             sg.In(str(Params.stride), key="-stride-", size=(5, 1))],
            [sg.Text("FPS fixed:"),
             sg.In("None", key="-fpsfixed-", size=(5, 1))],
            [sg.Text("            ")],
            [sg.Text("Skin Extractor:"),
             sg.Radio("Convex Hull", "RADIO2",
                      default=True, key="-skinConvex-"),
             sg.Radio("FaceParsing", "RADIO2", default=False, key="-skinFaceParsing-")],
            [sg.Text("ROI:"),
             sg.Radio("Holistic", "RADIO3",
                      default=False, key="-holistic-"),
             sg.Radio("Patches", "RADIO3", default=True, key="-patches-")],
            [sg.Text("Method for Signal computation:"),
             sg.Radio("mean", "RADIO0",
                      default=True, key="-mean-"),
             sg.Radio("median", "RADIO0", default=False, key="-median-")],
            [sg.Text("            ")],
            [sg.Text("Patches type:"),
             sg.Radio("Squares", "RADIO4",
                      default=True, key="-squares-"),
             sg.Radio("Rectangles", "RADIO4", default=False, key="-rects-")],
            [sg.Text("Landmarks list (ex: [1,2,3]):")],
            [sg.In("Default", key="-ldmkslist-")],
            [sg.Text("Squares dimension:"),
             sg.In(str(Params.squares_dim), key="-squares_dim-", size=(5, 1))],
            [sg.Text("Rects dimension (ex for two landmarks: [[20,15],[10,10]]):")],
            [sg.In("[[],]", key="-rectsdim-")],
            [sg.Text("            ")],
            [sg.Button('Apply Parameters', key='-APPLY-')],
            [sg.Button('START', key='-START-', visible=False)]
            ]
    col2 = [[sg.Text("            ")],
            [sg.Text("Skin RGB color low threshold (ex: input V -> RGB(V,V,V)):       "),
             sg.In(str(Params.skin_color_low_threshold), key="-skin_color_low_threshold-", size=(5, 1))],
            [sg.Text("Skin RGB color high threshold (ex: input V -> RGB(V,V,V)):      "),
             sg.In(str(Params.skin_color_high_threshold), key="-skin_color_high_threshold-", size=(5, 1))],
            [sg.Text("ROI RGB color low threshold (ex: input V -> RGB(V,V,V)):        "),
             sg.In(str(Params.sig_color_low_threshold), key="-sig_color_low_threshold-", size=(5, 1))],
            [sg.Text("ROI RGB color high threshold (ex: input V -> RGB(V,V,V)):       "),
             sg.In(str(Params.sig_color_high_threshold), key="-sig_color_high_threshold-", size=(5, 1))],
            [sg.Text("            ")],
            [sg.Text("Signal RGB color low threshold (ex: input V -> RGB(V,V,V)):     "),
             sg.In(str(Params.color_low_threshold), key="-color_low_threshold-", size=(5, 1))],
            [sg.Text("Signal RGB color high threshold (ex: input V -> RGB(V,V,V)):    "),
             sg.In(str(Params.color_high_threshold), key="-color_high_threshold-", size=(5, 1))],
            [sg.Text("            ")],
            [sg.Text("Visualize Skin:"),
             sg.Radio("True", "RADIO5",
                      default=True, key="-visualizeskintrue-"),
             sg.Radio("False", "RADIO5", default=False, key="-visualizeskinfalse-")],
            [sg.Text("Visualize Landmarks:"),
             sg.Radio("True", "RADIO6",
                      default=True, key="-visualizeldmkstrue-"),
             sg.Radio("False", "RADIO6", default=False, key="-visualizeldmksfalse-")],
            [sg.Text("\t - Visualize Patches:"),
             sg.Radio("True", "RADIO7",
                      default=True, key="-visualizepatchestrue-"),
             sg.Radio("False", "RADIO7", default=False, key="-visualizepatchesfalse-")],
            [sg.Text("\t - Visualize Landmarks number:"),
             sg.Radio("True", "RADIO8",
                      default=True, key="-visualizeldmksnumtrue-"),
             sg.Radio("False", "RADIO8", default=False, key="-visualizeldmksnumfalse-")],
            [sg.Text("            \n Font Color (R,G,B,A)")],
            [sg.In(str(Params.font_color), key='-FontColor-', size=(25, 1))],
            [sg.Text("Font Size: "), sg.In(
                str(Params.font_size), key='-FontSize-', size=(4, 1))],
            [sg.Text("For Editing PRE-POST filters, BVP method and BPM params edit:\n Params.pre_filter, Params.method, Params.post_filter,\n Params.minHz, Params.maxHz")]
            ]
    layout = [[sg.Column(col1, element_justification='l'),
              sg.Column(col2, element_justification='l')]]
    window = sg.Window('pyVHR', layout, finalize=True, location=(0, 0))

    bpm_plot = []
    bpm_save = []
    bpm_plot_max = 60
    image = None
    original_image = None
    skin_image = None
    patches_image = None
    vhr_t = None
    sd = SharedData()

    def create_run_window():
        fps = Params.fps_fixed if Params.fps_fixed is not None else get_fps(
            Params.videoFileName)
        visualize_list = [[sg.Radio(
            'Original Video',  "RADIO11", default=True, key="-IN_Visualize_original-")], ]
        if Params.visualize_skin:
            visualize_list.append(
                [sg.Radio('Skin', "RADIO11", default=False, key="-IN_Visualize_skin-")])
        if Params.approach == 'patches':
            visualize_list.append(
                [sg.Radio('Patches', "RADIO11", default=False, key="-IN_Visualize_patches-")])

        run_col1 = [[sg.Text(Params.videoFileName)],
                    [sg.Text(
                        "Video FPS: "+str(fps)),
                    sg.Text(
                        "Video resolution:             ", key="-res-")],
                    [sg.Text("Visualize: ")],
                    *visualize_list,
                    [sg.Text('BPMs')],
                    [sg.Image(filename='', key='-BPM_PLOT-')],
                    [sg.Button('Stop')],
                    [sg.Button('Save BPMs')]]
        run_col2 = [[sg.Text('Visualize')],
                    [sg.Image(filename='', key='-VIDEO-')]]
        run_layout = [[sg.Column(run_col1, element_justification='l'),
                       sg.Column(run_col2, element_justification='c')]]
        wr = sg.Window('pyVHR - RUN', run_layout,
                       finalize=True, location=(0, 0))
        # blank plots for compiling kernels
        imgbytes = cv2.imencode('.png', cv2.cvtColor(
            np.zeros((480, 640, 3), dtype=np.uint8), cv2.COLOR_BGR2RGB))[1].tobytes()
        wr['-VIDEO-'].update(data=imgbytes)
        fig = go.Figure()
        fig.add_trace(go.Scatter(x=np.arange(0, 10, 1), y=np.zeros(9),
                                 mode='lines+markers',
                                 name='lines+markers'))
        fig.update_layout(autosize=False, width=400, height=300, margin=dict(
            l=1, r=1, b=1, t=1, pad=1), paper_bgcolor="LightSteelBlue",)
        img_bytes = fig.to_image(format="png")
        wr['-BPM_PLOT-'].update(data=img_bytes)
        return wr

    run_window = None
    while True:
        win, event, values = sg.read_all_windows(timeout=16)
        if event in (None, 'Exit', 'Stop', sg.WIN_CLOSED) and run_window is not None:
            sd.q_stop_cap.put(0)  # stop cap, it will stop VHR model thread
            sd.q_stop.put(0)  # stop VHR model thread if cap can't
            vhr_t.join()
            vhr_t = None
            sd = None
            sd = SharedData()
            run_window.close()
            run_window = None
        elif event in (None, 'Exit', sg.WIN_CLOSED) and run_window is None:
            window.close()
            break
        if run_window is not None:
            run_win_event, run_win_values = run_window.read(timeout=16)
            # Video plot
            if not sd.q_video_image.empty():
                original_image = sd.q_video_image.get()
            if not sd.q_skin_image.empty():
                skin_image = sd.q_skin_image.get()
            if not sd.q_patches_image.empty():
                patches_image = sd.q_patches_image.get()
            if original_image is not None and bool(run_win_values["-IN_Visualize_original-"]):
                image = original_image
            elif Params.visualize_skin and skin_image is not None and bool(run_win_values["-IN_Visualize_skin-"]):
                image = skin_image
            elif Params.visualize_landmarks and patches_image is not None and bool(run_win_values["-IN_Visualize_patches-"]):
                image = patches_image
            if image is not None:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                imgbytes = cv2.imencode('.png', image)[1].tobytes()
                run_window['-VIDEO-'].update(data=imgbytes)
                run_window['-res-'].update("Video resolution: " +
                                           str(image.shape[1])+" x "+str(image.shape[0]))
            # BPM plot
            if not sd.q_bpm.empty():
                bpm = sd.q_bpm.get()
                bpm_plot.append(bpm)
                bpm_save.append(bpm)
                if len(bpm_plot) >= bpm_plot_max:
                    bpm_plot = bpm_plot[1:]
                fig = go.Figure()
                fig.add_trace(go.Scatter(x=np.arange(0, len(bpm_plot), Params.stride), y=np.array(bpm_plot),
                                         mode='lines+markers',
                                         name='lines+markers'))
                fig.update_layout(
                    autosize=False,
                    width=600,
                    height=400,
                    margin=dict(l=1, r=1, b=1, t=1, pad=1),
                    paper_bgcolor="LightSteelBlue",
                )
                img_bytes = fig.to_image(format="png")
                run_window['-BPM_PLOT-'].update(data=img_bytes)
            if run_win_event == 'Save BPMs':
                try:
                    path = Params.out_path if Params.out_path is not None else os.getcwd() + \
                        os.path.sep
                    with open(str(path) + str(time.time())+'_BPMs', 'wb') as f:
                        pickle.dump(bpm_save, f)
                    print("BPMs saved!")
                except:
                    print("[ERROR] wrong path for saving BPMs!")
        if event == '-APPLY-':
            if values['-VideoFileName-'] == '':
                Params.videoFileName = 0
                window['-VideoFileName-'].update('0')
            else:
                Params.videoFileName = values['-VideoFileName-']
            if values['-VideoFileName-'].isnumeric():
                Params.videoFileName = int(values['-VideoFileName-'])
                Params.fake_delay = False
            else:
                Params.fake_delay = True
            if values['-winSize-'].isnumeric():
                Params.winSize = int(
                    values['-winSize-']) if int(values['-winSize-']) > 1 else 3
            window['-winSize-'].update(str(Params.winSize))
            if values['-stride-'].isnumeric():
                Params.stride = int(
                    values['-stride-']) if int(values['-stride-']) > 0 else 1
            window['-stride-'].update(str(Params.stride))
            Params.cuda = bool(values['-cudaTrue-']
                               ) if bool(values['-cudaTrue-']) else False
            if values['-fpsfixed-'].isnumeric():
                Params.fps_fixed = int(
                    values['-fpsfixed-']) if int(values['-fpsfixed-']) > 0 else None
            window['-fpsfixed-'].update(str(Params.fps_fixed))
            Params.skin_extractor = 'convexhull' if bool(
                values['-skinConvex-']) else 'faceparsing'
            Params.approach = 'holistic' if bool(
                values['-holistic-']) else 'patches'
            Params.type = 'mean' if bool(
                values['-mean-']) else 'median'
            Params.patches = 'squares' if bool(
                values['-squares-']) else 'rects'
            try:
                ldlist = ast.literal_eval(values['-ldmkslist-'])
                Params.landmarks_list = ldlist if len(
                    ldlist) > 0 else Params.landmarks_list
            except:
                window['-ldmkslist-'].update(str(Params.landmarks_list))
            if values['-squares_dim-'].isnumeric():
                Params.squares_dim = float(
                    values['-squares_dim-']) if float(values['-squares_dim-']) > 0 else None
            else:
                window['-squares_dim-'].update(str(Params.squares_dim))
            try:
                Params.rects_dims = ast.literal_eval(values['-rectsdim-'])
                # Default if len is not the same
                if len(Params.rects_dims) != len(Params.landmarks_list):
                    new_rect_dim = []
                    for i in range(len(Params.landmarks_list)):
                        new_rect_dim.append(
                            [Params.squares_dim, Params.squares_dim])
                    Params.rects_dims = new_rect_dim
                    window['-rectsdim-'].update(str(new_rect_dim))
            except:
                # Default if parameter is wrong
                new_rect_dim = []
                for i in range(len(Params.landmarks_list)):
                    new_rect_dim.append(
                        [Params.squares_dim, Params.squares_dim])
                Params.rects_dims = new_rect_dim
                window['-rectsdim-'].update(str(new_rect_dim))

            if values['-skin_color_low_threshold-'].isnumeric():
                Params.skin_color_low_threshold = int(
                    values['-skin_color_low_threshold-']) if int(values['-skin_color_low_threshold-']) >= 0 and int(values['-skin_color_low_threshold-']) <= 255 else 2
            window['-skin_color_low_threshold-'].update(
                str(Params.skin_color_low_threshold))

            if values['-skin_color_high_threshold-'].isnumeric():
                Params.skin_color_high_threshold = int(
                    values['-skin_color_high_threshold-']) if int(values['-skin_color_high_threshold-']) >= 0 and int(values['-skin_color_high_threshold-']) <= 255 else 254
            window['-skin_color_high_threshold-'].update(
                str(Params.skin_color_high_threshold))

            if values['-sig_color_low_threshold-'].isnumeric():
                Params.sig_color_low_threshold = int(
                    values['-sig_color_low_threshold-']) if int(values['-sig_color_low_threshold-']) >= 0 and int(values['-sig_color_low_threshold-']) <= 255 else 2
            window['-sig_color_low_threshold-'].update(
                str(Params.sig_color_low_threshold))

            if values['-sig_color_high_threshold-'].isnumeric():
                Params.sig_color_high_threshold = int(
                    values['-sig_color_high_threshold-']) if int(values['-sig_color_high_threshold-']) >= 0 and int(values['-sig_color_high_threshold-']) <= 255 else 254
            window['-sig_color_high_threshold-'].update(
                str(Params.sig_color_high_threshold))

            if values['-color_low_threshold-'].isnumeric():
                Params.color_low_threshold = int(
                    values['-color_low_threshold-']) if int(values['-color_low_threshold-']) >= 0 and int(values['-color_low_threshold-']) <= 255 else -1
            window['-color_low_threshold-'].update(
                str(Params.color_low_threshold))

            if values['-color_high_threshold-'].isnumeric():
                Params.color_high_threshold = int(
                    values['-color_high_threshold-']) if int(values['-color_high_threshold-']) >= 0 and int(values['-color_high_threshold-']) <= 255 else 255
            window['-color_high_threshold-'].update(
                str(Params.color_high_threshold))

            Params.visualize_skin = True if bool(
                values['-visualizeskintrue-']) else False
            Params.visualize_landmarks = True if bool(
                values['-visualizeldmkstrue-']) else False
            Params.visualize_patches = True if bool(
                values['-visualizepatchestrue-']) else False
            Params.visualize_landmarks_number = True if bool(
                values['-visualizeldmksnumtrue-']) else False

            try:
                colorfont = ast.literal_eval(values['-FontColor-'])
                if len(colorfont) == 4:
                    correct = True
                    for e in colorfont:
                        if not(e >= 0 and e <= 255):
                            correct = False
                    if correct:
                        Params.font_color = colorfont
            except:
                pass
            window['-FontColor-'].update(str(Params.font_color))

            if values['-FontSize-'].isnumeric():
                Params.font_size = float(
                    values['-FontSize-']) if float(values['-FontSize-']) > 0.0 else 0.3
            window['-FontSize-'].update(str(Params.font_size))
            window['-START-'].update(visible=True)
        if event == '-START-' and not run_window:
            bpm_plot = []
            bpm_save = []
            image = None
            sd = SharedData()
            run_window = create_run_window()
            if vhr_t == None:
                vhr_t = threading.Thread(target=VHRroutine, args=(sd,))
                vhr_t.daemon = False
                vhr_t.start()
    window.close()


if __name__ == '__main__':
    # Theese parameters can be edited only here.

    # Pre and Post filters are list of dictionary with the following structure:
    # {'filter_func': method_name, 'params': {}}

    # A rPPG method is a dictionary with the following structure:
    # {'method_func': method_name, 'device_type': 'device_name', 'params': {}}
    # device_type can be: 'cpu', 'cuda', 'torch'.

    Params.pre_filter = [{'filter_func': BPfilter, 'params': {
        'minHz': 0.7, 'maxHz': 3.0, 'fps': 'adaptive', 'order': 6}}]

    Params.method = {'method_func': cupy_CHROM,
                     'device_type': 'cuda', 'params': {}}

    Params.post_filter = [{'filter_func': BPfilter, 'params': {
        'minHz': 0.7, 'maxHz': 3.0, 'fps': 'adaptive', 'order': 6}}]

    # BPM extraction type
    # 'psd_clustering' or welch
    Params.BPM_extraction_type = 'welch'

    # Saving path for BPMs
    #Params.out_path = "/home/user/Downloads/"
    Params.out_path = None  # Working directory will be used

    # Downscale input video to width 640 (preserve aspect-ratio)
    Params.resize = False

    GUI_MENU()



import plotly.graph_objects as go
import numpy as np

# Print iterations progress
def printProgressBar (iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = printEnd)
    # Print New Line on Complete
    if iteration == total: 
        print()
        
        
def multiplot(x=None, y=None, name=None, zeroMean=True, title="Signal", height=400, width=800):

    fig = go.Figure()

    if not np.any(y):
        return
    else:
        if y.ndim == 1:
            c = 1
            n = y.shape[0]
            if zeroMean:
                z = y-y.mean()
            if not np.any(x):
                x = np.linspace(0,n-1,n)
            fig.add_trace(go.Scatter(x=x,y=z,name=name))
        else:
            c,n = y.shape
            if not np.any(x):
                x = np.linspace(0,n-1,n)
            for i in range(c):
                z = y[i] 
                if name:
                    s = name[i]
                else:
                    s = "sig" + str(i)
                if zeroMean:
                    z = y[i]-y[i].mean()
                    
                fig.add_trace(go.Scatter(x=x,y=z,name=s))
    
    fig.update_layout(height=height, width=width, title=title,
        font=dict(
            family="Courier New, monospace",
            size=14,
            color="#7f7f7f")
    )

    fig.show()
    
    return fig


from numba import njit, prange, float32
import math
import time
import numpy as np
import PIL.Image
import torchvision.transforms as transforms
from numba import prange, njit
import os
import matplotlib.pyplot as plt


"""
This module defines classes or methods used for signal extraction.
"""

class SignalProcessingParams():
    """
        This class contains usefull parameters used by this module.

        RGB_LOW_TH (numpy.int32): RGB low-threshold value.

        RGB_HIGH_TH (numpy.int32): RGB high-threshold value.
    """
    RGB_LOW_TH = np.int32(55)
    RGB_HIGH_TH = np.int32(200)


@njit(['float32[:,:](uint8[:,:,:], int32, int32)', ], parallel=True, fastmath=True, nogil=True)
def holistic_mean(im, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method computes the RGB-Mean Signal excluding 'im' pixels
    that are outside the RGB range [RGB_LOW_TH, RGB_HIGH_TH] (extremes are included).

    Args: 
        im (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
        RGB_LOW_TH (numpy.int32): RGB low threshold value.
        RGB_HIGH_TH (numpy.int32): RGB high threshold value.
    
    Returns:
        RGB-Mean Signal as float32 ndarray with shape [1,3], where 1 is the single estimator,
        and 3 are r-mean, g-mean and b-mean.
    """
    mean = np.zeros((1, 3), dtype=np.float32)
    mean_r = np.float32(0.0)
    mean_g = np.float32(0.0)
    mean_b = np.float32(0.0)
    num_elems = np.float32(0.0)
    for x in prange(im.shape[0]):
        for y in prange(im.shape[1]):
            if not((im[x, y, 0] <= RGB_LOW_TH and im[x, y, 1] <= RGB_LOW_TH and im[x, y, 2] <= RGB_LOW_TH)
                    or (im[x, y, 0] >= RGB_HIGH_TH and im[x, y, 1] >= RGB_HIGH_TH and im[x, y, 2] >= RGB_HIGH_TH)):
                mean_r += im[x, y, 0]
                mean_g += im[x, y, 1]
                mean_b += im[x, y, 2]
                num_elems += 1.0
    if num_elems > 1.0:
        mean[0, 0] = mean_r / num_elems
        mean[0, 1] = mean_g / num_elems
        mean[0, 2] = mean_b / num_elems
    else:
        mean[0, 0] = mean_r
        mean[0, 1] = mean_g
        mean[0, 2] = mean_b 
    return mean


@njit(['float32[:,:](float32[:,:],uint8[:,:,:],float32, int32, int32)', ], parallel=True, fastmath=True, nogil=True)
def landmarks_mean(ldmks, im, square, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method computes the RGB-Mean Signal excluding 'im' pixels
    that are outside the RGB range [RGB_LOW_TH, RGB_HIGH_TH] (extremes are included).

    Args: 
        ldmks (float32 ndarray): landmakrs as ndarray with shape [num_landmarks, 5],
             where the second dimension contains y-coord, x-coord, r-mean (value is not important), g-mean (value is not important), b-mean (value is not important).
        im (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
        square (numpy.float32): side size of square patches.
        RGB_LOW_TH (numpy.int32): RGB low threshold value.
        RGB_HIGH_TH (numpy.int32): RGB high threshold value.
    
    Returns:
        RGB-Mean Signal as float32 ndarray with shape [num_landmarks, 5], where the second dimension contains y-coord, x-coord, r-mean, g-mean, b-mean.
    """
    r_ldmks = ldmks.astype(np.float32)
    width = im.shape[1]
    height = im.shape[0]
    S = math.floor(square/2)
    lds_mean = np.zeros((ldmks.shape[0], 3), dtype=np.float32)
    num_elems = np.zeros((ldmks.shape[0], ), dtype=np.float32)
    for ld_id in prange(0, r_ldmks.shape[0]):
        if r_ldmks[ld_id, 0] >= 0.0:
            for x in prange(int(r_ldmks[ld_id, 0] - S), int(r_ldmks[ld_id, 0] + S + 1)):
                for y in prange(int(r_ldmks[ld_id, 1] - S), int(r_ldmks[ld_id, 1] + S + 1)):
                    if x >= 0 and x < height and y >= 0 and y < width:
                        if not((im[x, y, 0] <= RGB_LOW_TH and im[x, y, 1] <= RGB_LOW_TH and im[x, y, 2] <= RGB_LOW_TH) or
                               (im[x, y, 0] >= RGB_HIGH_TH and im[x, y, 1] >= RGB_HIGH_TH and im[x, y, 2] >= RGB_HIGH_TH)):
                            lds_mean[ld_id, 0] += np.float32(im[x, y, 0])
                            lds_mean[ld_id, 1] += np.float32(im[x, y, 1])
                            lds_mean[ld_id, 2] += np.float32(im[x, y, 2])
                            num_elems[ld_id] += 1.0
            if num_elems[ld_id] > 1.0:
                r_ldmks[ld_id, 2] = lds_mean[ld_id, 0] / num_elems[ld_id]
                r_ldmks[ld_id, 3] = lds_mean[ld_id, 1] / num_elems[ld_id]
                r_ldmks[ld_id, 4] = lds_mean[ld_id, 2] / num_elems[ld_id]
    return r_ldmks


@njit(['float32[:,:](float32[:,:],uint8[:,:,:],float32, int32, int32)', ], parallel=True, fastmath=True, nogil=True)
def landmarks_median(ldmks, im, square, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method computes the RGB-Median Signal excluding 'im' pixels
    that are outside the RGB range [RGB_LOW_TH, RGB_HIGH_TH] (extremes are included).

    Args: 
        ldmks (float32 ndarray): landmakrs as ndarray with shape [num_landmarks, 5],
             where the second dimension contains y-coord, x-coord, r-mean (value is not important), g-mean (value is not important), b-mean (value is not important).
        im (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
        square (numpy.float32): side size of square patches.
        RGB_LOW_TH (numpy.int32): RGB low threshold value.
        RGB_HIGH_TH (numpy.int32): RGB high threshold value.
    
    Returns:
        RGB-Median Signal as float32 ndarray with shape [num_landmarks, 5], where the second dimension contains y-coord, x-coord, r-mean, g-mean, b-mean.
    """
    r_ldmks = ldmks.astype(np.float32)
    width = float32(im.shape[1])
    height = float32(im.shape[0])
    S = math.floor(square/2.0)
    for ld_id in prange(r_ldmks.shape[0]):
        if r_ldmks[ld_id, 0] >= 0.0:
            x_s = r_ldmks[ld_id, 0] - S
            if x_s < 0.0:
                x_s = 0
            x_e = r_ldmks[ld_id, 0] + S
            if x_e >= height:
                x_e = height
            y_s = r_ldmks[ld_id, 1] - S
            if y_s < 0.0:
                y_s = 0
            y_e = r_ldmks[ld_id, 1] + S
            if y_e >= width:
                y_e = width
            ar = np.copy(im[int(x_s):int(x_e), int(y_s):int(y_e), :])
            f_ar = ar.flatten()
            r = ar[:, :, 0].flatten()
            g = ar[:, :, 1].flatten()
            b = ar[:, :, 2].flatten()
            goodidx = np.ones((ar.shape[0]*ar.shape[1],), dtype=np.int32)
            targets = np.arange(3, f_ar.shape[0], 3)
            for idx in prange(targets.shape[0]):
                i = targets[idx]
                if ((f_ar[i-2] <= RGB_LOW_TH and f_ar[i-1] <= RGB_LOW_TH and f_ar[i] <= RGB_LOW_TH) or
                        (f_ar[i-2] >= RGB_HIGH_TH and f_ar[i-1] >= RGB_HIGH_TH and f_ar[i] >= RGB_HIGH_TH)):
                    goodidx[i % 3] = 0
            goodidx = np.argwhere(goodidx).flatten()
            if goodidx.size < 1 or r.size < 1:
                r_ldmks[ld_id, 2] = np.float32(0.0)
                r_ldmks[ld_id, 3] = np.float32(0.0)
                r_ldmks[ld_id, 4] = np.float32(0.0)
            else:
                r_ldmks[ld_id, 2] = np.float32(np.median(r[goodidx]))
                r_ldmks[ld_id, 3] = np.float32(np.median(g[goodidx]))
                r_ldmks[ld_id, 4] = np.float32(np.median(b[goodidx]))
    return r_ldmks


@njit(['float32[:,:](float32[:,:],uint8[:,:,:],float32[:,:], int32, int32)', ], parallel=True, fastmath=True, nogil=True)
def landmarks_mean_custom_rect(ldmks, im, rects, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method computes the RGB-Mean Signal excluding 'im' pixels
    that are outside the RGB range [RGB_LOW_TH, RGB_HIGH_TH] (extremes are included).

    Args: 
        ldmks (float32 ndarray): landmakrs as ndarray with shape [num_landmarks, 5],
             where the second dimension contains y-coord, x-coord, r-mean (value is not important), g-mean (value is not important), b-mean (value is not important).
        im (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
        rects (float32 ndarray): positive float32 np.ndarray of shape [num_landmarks, 2]. If the list of used landmarks is [1,2,3] 
            and rects_dim is [[10,20],[12,13],[40,40]] then the landmark number 2 will have a rectangular patch of xy-dimension 12x13.
        RGB_LOW_TH (numpy.int32): RGB low threshold value.
        RGB_HIGH_TH (numpy.int32): RGB high threshold value.
    
    Returns:
        RGB-Mean Signal as float32 ndarray with shape [num_landmarks, 5], where the second dimension contains y-coord, x-coord, r-mean, g-mean, b-mean.
    """
    r_ldmks = ldmks.astype(np.float32)
    width = im.shape[1]
    height = im.shape[0]
    lds_mean = np.zeros((ldmks.shape[0], 3), dtype=np.float32)
    num_elems = np.zeros((ldmks.shape[0], ), dtype=np.float32)
    for ld_id in prange(0, r_ldmks.shape[0]):
        if r_ldmks[ld_id, 0] >= 0:
            Sx = math.floor(rects[ld_id, 1]/2)
            Sy = math.floor(rects[ld_id, 0]/2)
            for x in prange(r_ldmks[ld_id, 0] - Sx, r_ldmks[ld_id, 0] + Sx + 1):
                for y in prange(r_ldmks[ld_id, 1] - Sy, r_ldmks[ld_id, 1] + Sy + 1):
                    if x >= 0 and x < height and y >= 0 and y < width:
                        if not((im[x, y, 0] <= RGB_LOW_TH and im[x, y, 1] <= RGB_LOW_TH and im[x, y, 2] <= RGB_LOW_TH) or
                               (im[x, y, 0] >= RGB_HIGH_TH and im[x, y, 1] >= RGB_HIGH_TH and im[x, y, 2] >= RGB_HIGH_TH)):
                            lds_mean[ld_id, 0] += im[x, y, 0]
                            lds_mean[ld_id, 1] += im[x, y, 1]
                            lds_mean[ld_id, 2] += im[x, y, 2]
                            num_elems[ld_id] += 1.0
            if num_elems[ld_id] > 1.0:
                r_ldmks[ld_id, 2] = lds_mean[ld_id, 0] / num_elems[ld_id]
                r_ldmks[ld_id, 3] = lds_mean[ld_id, 1] / num_elems[ld_id]
                r_ldmks[ld_id, 4] = lds_mean[ld_id, 2] / num_elems[ld_id]
    return r_ldmks


@njit(['float32[:,:](float32[:,:],uint8[:,:,:],float32[:,:], int32, int32)', ], parallel=True, fastmath=True, nogil=True)
def landmarks_median_custom_rect(ldmks, im, rects, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method computes the RGB-Median Signal excluding 'im' pixels
    that are outside the RGB range [RGB_LOW_TH, RGB_HIGH_TH] (extremes are included).

    Args: 
        ldmks (float32 ndarray): landmakrs as ndarray with shape [num_landmarks, 5],
             where the second dimension contains y-coord, x-coord, r-mean (value is not important), g-mean (value is not important), b-mean (value is not important).
        im (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
        rects (float32 ndarray): positive float32 np.ndarray of shape [num_landmarks, 2]. If the list of used landmarks is [1,2,3] 
            and rects_dim is [[10,20],[12,13],[40,40]] then the landmark number 2 will have a rectangular patch of xy-dimension 12x13.
        RGB_LOW_TH (numpy.int32): RGB low threshold value.
        RGB_HIGH_TH (numpy.int32): RGB high threshold value.
    
    Returns:
        RGB-Median Signal as float32 ndarray with shape [num_landmarks, 5], where the second dimension contains y-coord, x-coord, r-mean, g-mean, b-mean.
    """
    r_ldmks = ldmks.astype(np.float32)
    width = float32(im.shape[1])
    height = float32(im.shape[0])
    for ld_id in prange(ldmks.shape[0]):
        if r_ldmks[ld_id, 0] >= 0.0:
            Sx = math.floor(rects[ld_id, 1]/2)
            Sy = math.floor(rects[ld_id, 0]/2)
            x_s = r_ldmks[ld_id, 0] - Sx
            if x_s < 0.0:
                x_s = 0
            x_e = r_ldmks[ld_id, 0] + Sx
            if x_e >= height:
                x_e = height
            y_s = r_ldmks[ld_id, 1] - Sy
            if y_s < 0.0:
                y_s = 0
            y_e = r_ldmks[ld_id, 1] + Sy
            if y_e >= width:
                y_e = width
            ar = np.copy(im[int(x_s):int(x_e), int(y_s):int(y_e), :])
            f_ar = ar.flatten()
            r = ar[:, :, 0].flatten()
            g = ar[:, :, 1].flatten()
            b = ar[:, :, 2].flatten()
            goodidx = np.ones((ar.shape[0]*ar.shape[1],), dtype=np.int32)
            targets = np.arange(3, f_ar.shape[0], 3)
            for idx in prange(targets.shape[0]):
                i = targets[idx]
                if ((f_ar[i-2] <= RGB_LOW_TH and f_ar[i-1] <= RGB_LOW_TH and f_ar[i] <= RGB_LOW_TH) or
                        (f_ar[i-2] >= RGB_HIGH_TH and f_ar[i-1] >= RGB_HIGH_TH and f_ar[i] >= RGB_HIGH_TH)):
                    goodidx[i % 3] = 0
            goodidx = np.argwhere(goodidx).flatten()
            if goodidx.size < 1 or r.size < 1:
                r_ldmks[ld_id, 2] = np.int32(0)
                r_ldmks[ld_id, 3] = np.int32(0)
                r_ldmks[ld_id, 4] = np.int32(0)
            else:
                r_ldmks[ld_id, 2] = np.int32(np.median(r[goodidx]))
                r_ldmks[ld_id, 3] = np.int32(np.median(g[goodidx]))
                r_ldmks[ld_id, 4] = np.int32(np.median(b[goodidx]))
    return r_ldmks



from pyVHR.extraction.utils import MagicLandmarks
from pyVHR.BVP.methods import *
from pyVHR.BVP.filters import *
from pyVHR.extraction.utils import MagicLandmarks

class Params:
    # DEFAULT PARAMETERS
    videoFileName = ''
    fps_fixed = None
    tot_sec = 0
    winSize = 6
    stride = 1
    cuda = True
    skin_extractor = 'convexhull'  # or faceparsing
    approach = 'patches'  # or holistic
    patches = 'squares'  # or rects
    type = 'mean'
    landmarks_list = MagicLandmarks.equispaced_facial_points
    squares_dim = 30.0
    rects_dims = []
    # extractor params
    skin_color_low_threshold = 75
    skin_color_high_threshold = 230
    sig_color_low_threshold = 75
    sig_color_high_threshold = 230
    # rgb_filter_th params
    color_low_threshold = 75
    color_high_threshold = 230
    # visualize skin and patches
    visualize_skin = True
    visualize_patches = True
    visualize_landmarks = True
    visualize_landmarks_number = True
    font_size = 0.3
    font_color = (255, 0, 0, 255)

    # Pre filtering
    # dictionary of {filter_func, params}
    pre_filter = [{'filter_func': BPfilter, 'params': {
        'minHz': 0.7, 'maxHz': 3.0, 'fps': 'adaptive', 'order': 6}}]

    # BVP method
    # dictionary of {method_func, device_type, params}
    method = {'method_func': cupy_CHROM,
              'device_type': 'cuda', 'params': {}}

    # Post filtering
    # dictionary of {filter_func, params}
    post_filter = [{'filter_func': BPfilter, 'params': {
        'minHz': 0.7, 'maxHz': 3.0, 'fps': 'adaptive', 'order': 6}}]

    # BPM params
    minHz = 0.7
    maxHz = 3.0
    # WELCH: CPU, GPU
    # PSD_CLUSTERING: CPU, GPU
    # USE psd_clustering only with patches!
    BPM_extraction_type = "welch" # or 'psd_clustering'


    # Utils
    fake_delay = False
    resize = True
    out_path = None



import pandas as pd
import numpy as np
import os
import re 
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import scipy.stats as ss
import scikit_posthocs as sp
import math
from autorank import autorank, plot_stats, create_report
from matplotlib.colors import ListedColormap
from matplotlib.colorbar import ColorbarBase, Colorbar


class StatAnalysis():
    """ 
    Statistic analyses for multiple datasets and multiple rPPG methods
    """
    
    def __init__(self, filepath, join_data=False, remove_ouliers=False):
        """
        Args:
            filepath:
                - The path to the file contaning the results to test
            join_data: 
                - 'True' - If filepath is a folder, join the dataframes contained in the folder (to be used when wanting to merge multiple results from the same pipeline on the same dataset)
                - 'False' - (default) To be used if you want to test the same pipeline (eventually with multiple methods) on multiple datasets
            remove_ouliers:
                - 'True' -  Remove outliers from data prior to statistical testing
                - 'False' - (default) no outlier removal
        """
        
        if os.path.isdir(filepath):  
            self.multidataset = True
            self.path = filepath + "/"
            self.datasetsList = os.listdir(filepath)
        elif os.path.isfile(filepath):  
            self.multidataset = False
            self.datasetsList = [filepath]
            self.path = ""
        else:  
            raise("Error: filepath is wrong!")

        self.join_data = join_data
        self.available_metrics = ['MAE','RMSE','PCC','CCC']
        self.remove_ouliers = remove_ouliers

        # -- get data
        self.__getMethods()
        self.metricSort = {'MAE':'min','RMSE':'min','PCC':'max', 'CCC': 'max'}
        self.scale = {'MAE':'log','RMSE':'log','PCC':'linear', 'CCC':'linear'}
        
        self.use_stats_pipeline = False

    def run_stats(self, methods=None, metric='CCC', approach='frequentist', print_report=True):
        """
        Runs the statistical testing procedure by automatically selecting the appropriate test for the available data.

        Args:
            methods:
                - The rPPG methods to analyze
            metric:
                - 'MAE' - Mean Absolute Error
                - 'RMSE' - Root Mean Squared Error
                - 'PCC' - Pearson's Correlation Coefficient
                - 'CCC' - Concordance Correlation Coefficient
            approach:
                - 'frequentist' - (default) Use frequentist hypotesis tests for the analysis
                - 'bayesian' - Use bayesian hypotesis tests for the analysis
            print_report:
                - 'True' - (default) print a report of the hypotesis testing procedure
                - 'False' - Doesn't print any report
        """
        metric = metric.upper()
        assert metric in self.available_metrics, 'Error! Available metrics are ' + str(self.available_metrics)
        
        # -- Method(s) 
        if methods != None:
            if not set(methods).issubset(set(self.methods)):
                raise ValueError("Some method is wrong!")
            else:
                self.methods = methods

        assert approach=='frequentist' or approach=='bayesian', "Approach should be 'frequentist' or bayesian, not " + str(approach)
           
        # -- set metric
        self.metric = metric
        self.mag = self.metricSort[metric]
        
        # -- get data from dataset(s)
        if self.multidataset:
            Y = self.__getData()
        else:
            Y = self.__getDataMono()
        self.ndataset = Y.shape[0]

        if metric == 'MAE':
            order='ascending'
        else:
            order='descending'

        Y_df = pd.DataFrame(Y, columns=self.methods)

        results = autorank(Y_df, alpha=0.05, order=order, verbose=False, approach=approach)
        self.stat_result = results
        self.use_stats_pipeline = True

        if approach=='bayesian':
            res_df = results.rankdf.iloc[:, [0,1,4,5,8]]
            print(res_df)

        if print_report:
            print(' ')
            create_report(results)
            print(' ')

        _ = self.computeCD(approach=approach)

        return Y_df
        
    
    def SignificancePlot(self, methods=None, metric='MAE'):
        """
        Plots the results of hypotesis testing with the significance plot 
        """

        # -- Method(s) 
        if methods == None:
            methods = self.methods
        else:
            if not set(methods).issubset(set(self.methods)):
                raise ValueError("Some method is wrong!")
            else:
                self.methods = methods
           
        # -- set metric
        self.metric = metric
        self.mag = self.metricSort[metric]
        
        # -- get data from dataset(s)
        if self.multidataset:
            Y = self.__getData()
        else:
            Y = self.__getDataMono()
        
        # -- Significance plot, a heatmap of p values
        methodNames = [x.upper() for x in self.methods]
        Ypd = pd.DataFrame(Y, columns=methodNames)
        ph = sp.posthoc_nemenyi_friedman(Ypd)
        cmap = ['1', '#fb6a4a',  '#08306b',  '#4292c6', '#c6dbef']
        heatmap_args = {'cmap': cmap, 'linewidths': 0.25, 'linecolor': '0.5', 
                        'clip_on': False, 'square': True, 'cbar_ax_bbox': [0.85, 0.35, 0.04, 0.3]}

        plt.figure(figsize=(5,4))
        sp.sign_plot(ph, cbar=True, **heatmap_args)
        plt.title('p-vals')
        
        fname = 'SP_' + self.metric + '.pdf'
        plt.savefig(fname)
        plt.show()
        
    def computeCD(self, avranks=None, numDatasets=None, alpha='0.05', display=True, approach='frequentist'):
        """
        Returns critical difference for Nemenyi or Bonferroni-Dunn test according 
        to given alpha (either alpha=”0.05” or alpha=”0.1”) for average ranks and 
        number of tested datasets N. Test can be either “nemenyi” for for Nemenyi 
        two tailed test or “bonferroni-dunn” for Bonferroni-Dunn test.
        See Orange package docs.
        """
        cd = self.stat_result.cd
        if display and approach=='frequentist':
            plot_stats(self.stat_result, allow_insignificant=True)
            plt.show()
        elif display and approach=='bayesian':
            self.plot_bayesian_res(self.stat_result)
            plt.show()
        return cd

    def plot_bayesian_res(self, stat_result):
        """
        Plots the results of bayesian significance testing
        """
        dm = stat_result.decision_matrix.copy()
        cmap = ['1', '#fb6a4a',  '#08306b',  '#4292c6']#, '#c6dbef']
        heatmap_args = {'cmap': cmap, 'linewidths': 0.25, 'linecolor': '0.5', 
                        'clip_on': False, 'square': True, 'cbar_ax_bbox': [0.85, 0.35, 0.04, 0.3]}
        dm[dm=='inconclusive'] = 0
        dm[dm=='smaller'] = 1
        dm[dm=='larger'] = 2
        np.fill_diagonal(dm.values, -1)

        pl, ax = plt.subplots()
        ax.imshow(dm.values.astype(int), cmap=ListedColormap(cmap))
        labels = list(dm.columns)
        # Major ticks
        ax.set_xticks(np.arange(0, len(labels), 1))
        ax.set_yticks(np.arange(0, len(labels), 1))
        # Labels for major ticks
        ax.set_xticklabels(labels, rotation='vertical')
        ax.set_yticklabels(labels)
        # Minor ticks
        ax.set_xticks(np.arange(-.5, len(labels), 1), minor=True)
        ax.set_yticks(np.arange(-.5, len(labels), 1), minor=True)
        ax.grid(which='minor', color='k', linestyle='-', linewidth=1)
        ax.set_title('Metric: ' + self.metric)
        cbar_ax = ax.figure.add_axes([0.85, 0.35, 0.04, 0.3])
        cbar = ColorbarBase(cbar_ax, cmap=ListedColormap(cmap), boundaries=[0, 1, 2, 3, 4])
        cbar.set_ticks(np.linspace(0.5, 3.5, 4))
        cbar.set_ticklabels(['None', 'equivalent', 'smaller', 'larger'])
        cbar.outline.set_linewidth(1)
        cbar.outline.set_edgecolor('0.5')
        cbar.ax.tick_params(size=0)

    def displayBoxPlot(self, methods=None, metric='MAE', scale=None, title=True):
        """
        Shows the distribution of populations with box-plots 
        """
        metric = metric.upper()

        # -- Method(s) 
        if methods == None:
            methods = self.methods
        else:
            if not set(methods).issubset(set(self.methods)):
                raise ValueError("Some method is wrong!")
            else:
                self.methods = methods
        
        if sorted(list(set(self.methods))) != sorted(self.methods):
            methods = self.datasetsList
        else:
            methods = self.methods

        # -- set metric
        self.metric = metric
        self.mag = self.metricSort[metric]
        if scale == None:
            scale = self.scale[metric]
        
        # -- get data from dataset(s)
        if self.multidataset:
            Y = self.__getData()
        else:
            Y = self.__getDataMono()
       
        # -- display box plot
        fig = self.boxPlot(methods, metric, Y, scale=scale, title=title)
        
        return fig

    def boxPlot(self, methods, metric, Y, scale, title):
        """
        Creates the box plot 
        """
        
        #  Y = mat(n-datasets,k-methods)   
        
        k = len(methods)
        
        if not (k == Y.shape[1]):
            raise("error!")

        offset = 50
        fig = go.Figure()

        methodNames = [x.upper() for x in methods]
        for i in range(k):
            yd = Y[:,i]
            name = methodNames[i]
            if len(np.argwhere(np.isnan(yd)).flatten())!=0:
                print(f"Warning! Video {self.dataFrame[0]['videoFilename'][np.argwhere(np.isnan(yd)).flatten()[0]]} contains NaN value for method {name}")
                continue
            # -- set color for box
            if metric == 'MAE' or  metric == 'RMSE' or metric == 'TIME_REQUIREMENT':
                med = np.median(yd)
                col = str(min(200,5*int(med)+offset))
            if metric == 'CC' or metric == 'PCC' or metric == 'CCC':
                med = 1-np.abs(np.median(yd))
                col = str(int(200*med)+offset)

            # -- add box 
            fig.add_trace(go.Box(
                y=yd,
                name=name,
                boxpoints='all',
                jitter=.7,
                #whiskerwidth=0.2,
                fillcolor="rgba("+col+","+col+","+col+",0.5)",
                line_color="rgba(0,0,255,0.5)",
                marker_size=2,
                line_width=2)
            )

        gwidth = np.max(Y)/10
        
        if title:
            tit = "Metric: " + metric
            top = 40
        else:
            tit=''
            top = 10
        
        fig.update_layout(
            title=tit,
            yaxis_type=scale,
            xaxis_type="category",
            yaxis=dict(
                autorange=True,
                showgrid=True,
                zeroline=True,
                #dtick=gwidth,
                gridcolor='rgb(255,255,255)',
                gridwidth=.1,
                zerolinewidth=2,
                titlefont=dict(size=30)
            ),
            font=dict(
                family="monospace",
                size=16,
                color='rgb(20,20,20)'
            ),
            margin=dict(
                l=20,
                r=10,
                b=20,
                t=top,
            ),
            paper_bgcolor='rgb(250, 250, 250)',
            plot_bgcolor='rgb(243, 243, 243)',
            showlegend=False
        )

        #fig.show()
        return fig
        
    def saveStatsData(self, methods=None, metric='MAE', outfilename='statsData.csv'):
        """
        Saves statistics of data on disk 
        """
        Y = self.getStatsData(methods=methods, metric=metric, printTable=False)
        np.savetxt(outfilename, Y)
        
    def getStatsData(self, methods=None, metric='MAE', printTable=True):
        """
        Computes statistics of data 
        """
         # -- Method(s) 
        if methods == None:
            methods = self.methods
        else:
            if set(methods) <= set(self.methods):
                raise("Some method is wrong!")
            else:
                self.methods = methods
           
        # -- set metric
        self.metric = metric
        self.mag = self.metricSort[metric]
        
          # -- get data from dataset(s)
        #    return Y = mat(n-datasets,k-methods)   
        if self.multidataset:
            Y = self.__getData()
        else:
            Y = self.__getDataMono()
        
        # -- add median and IQR
        I = ss.iqr(Y,axis=0)
        M = np.median(Y,axis=0)
        Y = np.vstack((Y,M))
        Y = np.vstack((Y,I))
        
        if printTable:
            methodNames = [x.upper() for x in self.methods]
            dataseNames = self.datasetNames
            dataseNames.append('Median')
            dataseNames.append('IQR')
            df = pd.DataFrame(Y, columns=methodNames, index=dataseNames)
            display(df)
        
        return Y

    def __remove_outliers(self, df, factor=3.5):
        """
        Removes the ouliers. A data point is considered an outlier if 
        lies outside factor times the inter-quartile range of the data distribution 
        """
        Q1 = df.quantile(0.25)
        Q3 = df.quantile(0.75)
        IQR = Q3 - Q1
        df_out = df[~((df < (Q1 - factor*IQR)) | (df > (Q3 + factor*IQR))).any(axis=1)]
        return df_out

    def __getDataMono(self):
        mag = self.mag
        metric = self.metric
        methods = self.methods
        
        frame = self.dataFrame[0]

        # -- loop on methods
        Y = []
        
        for method in methods:
            vals = frame[frame['method'] == method][metric]
            if mag == 'min':
                data = [v[np.argmin(v)] for v in vals]
            else:
                data = [v[np.argmax(v)] for v in vals]
            Y.append(data)

        if self.remove_ouliers:
            res = pd.DataFrame(np.array(Y).T)
            res = self.__remove_outliers(res)
            res = res.to_numpy()
        else:
            res = np.array(Y).T
            
        return res

    def __getData(self):
        
        mag = self.mag
        metric = self.metric
        methods = self.methods
        
        # -- loop on datasets
        Y = []
        m_list = []
        for i,frame in enumerate(self.dataFrame):
            # -- loop on methods
            y = []
            for method in methods:
                vals = frame[frame['method'] == method][metric]
                if vals.empty:
                    continue
                m_list.append(method)
                if mag == 'min':
                    data = [v[np.argmin(v)] for v in vals]
                else:
                    data = [v[np.argmax(v)] for v in vals]
                y.append(data)
            y = np.array(y)

            if not self.join_data:
                Y.append(np.mean(y,axis=1))
            else:
                Y.append(y.T)   

        if not self.join_data:
            res = np.array(Y)
        else:
            self.methods = m_list
            n_dpoints = [curr_y.shape[0] for curr_y in Y]
            if len(set(n_dpoints)) != 1:
                raise("There should be the exact same number of elements in each dataset to join when 'join_data=True'")
            res = np.hstack(Y)

        if self.remove_ouliers:
            res = pd.DataFrame(res)
            res = self.__remove_outliers(res)
            res = res.to_numpy()

        return res
    
    def __getMethods(self):
        
        mets = []
        dataFrame = []
        N = len(self.datasetsList)
        
        # -- load dataframes
        self.datasetNames = []
        for file in self.datasetsList:
            filename = self.path + file
            self.datasetNames.append(file)            
            data = pd.read_hdf(filename)

            mets.append(set(list(data['method'])))
            dataFrame.append(data)
        
        if not self.join_data:
            # -- method names intersection among datasets
            methods = set(mets[0])
            if N > 1:
                for m in range(1,N-1):
                    methods.intersection(mets[m])
            methods = list(methods)
        else:
            methods = sum([list(m) for m in mets], [])
            if sorted(list(set(methods))) != sorted(methods):
                raise("Found multiple methods with the same name... Please ensure using different names for each method when 'join_data=True'")

        methods.sort()
        self.methods = methods
        self.dataFrame = dataFrame


from numba import cuda
import torch
import os


def cuda_info():
    if torch.cuda.is_available():
        print("# CUDA devices: ", torch.cuda.device_count())
        for e in range(torch.cuda.device_count()):
            print("# device number ", e, ": ", torch.cuda.get_device_name(e))


def select_cuda_device(n):
    torch.cuda.device(n)
    cuda.select_device(n)



import csv
import numpy as np
from pyVHR.datasets.dataset import Dataset
from pyVHR.BPM.BPM import BVPsignal


class UBFC1(Dataset):
    """
    UBFC1 Dataset

    .. UBFC dataset structure:
    .. -----------------
    ..     datasetDIR/
    ..     |   |-- SubjDIR1/
    ..     |       |-- vid.avi
    ..     |...
    ..     |   |-- SubjDIRM/
    ..     |       |-- vid.avi
    """
    name = 'UBFC1'
    signalGT = 'BVP'     # GT signal type
    numLevels = 2        # depth of the filesystem collecting video and BVP files
    numSubjects = 8     # number of subjects
    video_EXT = 'avi'    # extension of the video files
    frameRate = 30       # vieo frame rate
    VIDEO_SUBSTRING = ''  # substring contained in the filename
    SIG_EXT = 'xmp'     # extension of the BVP files
    SIG_SUBSTRING = ''  # substring contained in the filename
    SIG_SampleRate = 62  # sample rate of the BVP files
    skinThresh = [40, 60]  # thresholds for skin detection

    def readSigfile(self, filename):
        """ 
        Load signal from file.
        
        Returns:
            a :py:class:`pyVHR.BPM.BPM.BVPsignal` object that can be used to extract BPM signal from ground truth BVP signal.
        """
        gtTrace = []
        gtTime = []
        gtHR = []
        with open(filename, 'r') as csvfile:
            xmp = csv.reader(csvfile)
            for row in xmp:
                gtTrace.append(float(row[3]))
                gtTime.append(float(row[0])/1000.)
                gtHR.append(float(row[1]))

        data = np.array(gtTrace)
        time = np.array(gtTime)
        hr = np.array(gtHR)
        self.SIG_SampleRate = np.round(1/np.mean(np.diff(time)))

        '''import matplotlib.pyplot as plt
        plt.plot(hr)
        plt.show()'''

        return BVPsignal(data, self.SIG_SampleRate)



import time

def tic(verb=True):
    """ tic like Matlab function """

    global tic_toc_time
    tic_toc_time = time.time()
    if verb:
        print('start...')
    return tic_toc_time

def toc(verb=True):
    """ toc like Matlab function """

    global tic_toc_time
    T1 = time.time()-tic_toc_time
    if verb:
        print('elapsed = ' + str(T1))
    return T1



#!/usr/bin/python
# -*- encoding: utf-8 -*-

from logger import setup_logger
from model import BiSeNet
from face_dataset import FaceMask

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.nn.functional as F
import torch.distributed as dist

import os
import os.path as osp
import logging
import time
import numpy as np
from tqdm import tqdm
import math
from PIL import Image
import torchvision.transforms as transforms
import cv2

def vis_parsing_maps(im, parsing_anno, stride, save_im=False, save_path='vis_results/parsing_map_on_im.jpg'):
    # Colors for all 20 parts
    part_colors = [[255, 0, 0], [255, 85, 0], [255, 170, 0],
                   [255, 0, 85], [255, 0, 170],
                   [0, 255, 0], [85, 255, 0], [170, 255, 0],
                   [0, 255, 85], [0, 255, 170],
                   [0, 0, 255], [85, 0, 255], [170, 0, 255],
                   [0, 85, 255], [0, 170, 255],
                   [255, 255, 0], [255, 255, 85], [255, 255, 170],
                   [255, 0, 255], [255, 85, 255], [255, 170, 255],
                   [0, 255, 255], [85, 255, 255], [170, 255, 255]]

    im = np.array(im)
    vis_im = im.copy().astype(np.uint8)
    vis_parsing_anno = parsing_anno.copy().astype(np.uint8)
    vis_parsing_anno = cv2.resize(vis_parsing_anno, None, fx=stride, fy=stride, interpolation=cv2.INTER_NEAREST)
    vis_parsing_anno_color = np.zeros((vis_parsing_anno.shape[0], vis_parsing_anno.shape[1], 3)) + 255

    num_of_class = np.max(vis_parsing_anno)

    for pi in range(1, num_of_class + 1):
        index = np.where(vis_parsing_anno == pi)
        vis_parsing_anno_color[index[0], index[1], :] = part_colors[pi]

    vis_parsing_anno_color = vis_parsing_anno_color.astype(np.uint8)
    # print(vis_parsing_anno_color.shape, vis_im.shape)
    vis_im = cv2.addWeighted(cv2.cvtColor(vis_im, cv2.COLOR_RGB2BGR), 0.4, vis_parsing_anno_color, 0.6, 0)

    # Save result or not
    if save_im:
        cv2.imwrite(save_path, vis_im, [int(cv2.IMWRITE_JPEG_QUALITY), 100])

    # return vis_im

def evaluate(respth='./res/test_res', dspth='./data', cp='model_final_diss.pth'):

    if not os.path.exists(respth):
        os.makedirs(respth)

    n_classes = 19
    net = BiSeNet(n_classes=n_classes)
    net.cuda()
    save_pth = osp.join('res/cp', cp)
    net.load_state_dict(torch.load(save_pth))
    net.eval()

    to_tensor = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
    ])
    with torch.no_grad():
        for image_path in os.listdir(dspth):
            img = Image.open(osp.join(dspth, image_path))
            image = img.resize((512, 512), Image.BILINEAR)
            img = to_tensor(image)
            img = torch.unsqueeze(img, 0)
            img = img.cuda()
            out = net(img)[0]
            parsing = out.squeeze(0).cpu().numpy().argmax(0)

            vis_parsing_maps(image, parsing, stride=1, save_im=True, save_path=osp.join(respth, image_path))







if __name__ == "__main__":
    setup_logger('./res')
    evaluate()



#!/usr/bin/python
# -*- encoding: utf-8 -*-


from PIL import Image
import PIL.ImageEnhance as ImageEnhance
import random
import numpy as np

class RandomCrop(object):
    def __init__(self, size, *args, **kwargs):
        self.size = size

    def __call__(self, im_lb):
        im = im_lb['im']
        lb = im_lb['lb']
        assert im.size == lb.size
        W, H = self.size
        w, h = im.size

        if (W, H) == (w, h): return dict(im=im, lb=lb)
        if w < W or h < H:
            scale = float(W) / w if w < h else float(H) / h
            w, h = int(scale * w + 1), int(scale * h + 1)
            im = im.resize((w, h), Image.BILINEAR)
            lb = lb.resize((w, h), Image.NEAREST)
        sw, sh = random.random() * (w - W), random.random() * (h - H)
        crop = int(sw), int(sh), int(sw) + W, int(sh) + H
        return dict(
                im = im.crop(crop),
                lb = lb.crop(crop)
                    )


class HorizontalFlip(object):
    def __init__(self, p=0.5, *args, **kwargs):
        self.p = p

    def __call__(self, im_lb):
        if random.random() > self.p:
            return im_lb
        else:
            im = im_lb['im']
            lb = im_lb['lb']

            # atts = [1 'skin', 2 'l_brow', 3 'r_brow', 4 'l_eye', 5 'r_eye', 6 'eye_g', 7 'l_ear', 8 'r_ear', 9 'ear_r',
            #         10 'nose', 11 'mouth', 12 'u_lip', 13 'l_lip', 14 'neck', 15 'neck_l', 16 'cloth', 17 'hair', 18 'hat']

            flip_lb = np.array(lb)
            flip_lb[lb == 2] = 3
            flip_lb[lb == 3] = 2
            flip_lb[lb == 4] = 5
            flip_lb[lb == 5] = 4
            flip_lb[lb == 7] = 8
            flip_lb[lb == 8] = 7
            flip_lb = Image.fromarray(flip_lb)
            return dict(im = im.transpose(Image.FLIP_LEFT_RIGHT),
                        lb = flip_lb.transpose(Image.FLIP_LEFT_RIGHT),
                    )


class RandomScale(object):
    def __init__(self, scales=(1, ), *args, **kwargs):
        self.scales = scales

    def __call__(self, im_lb):
        im = im_lb['im']
        lb = im_lb['lb']
        W, H = im.size
        scale = random.choice(self.scales)
        w, h = int(W * scale), int(H * scale)
        return dict(im = im.resize((w, h), Image.BILINEAR),
                    lb = lb.resize((w, h), Image.NEAREST),
                )


class ColorJitter(object):
    def __init__(self, brightness=None, contrast=None, saturation=None, *args, **kwargs):
        if not brightness is None and brightness>0:
            self.brightness = [max(1-brightness, 0), 1+brightness]
        if not contrast is None and contrast>0:
            self.contrast = [max(1-contrast, 0), 1+contrast]
        if not saturation is None and saturation>0:
            self.saturation = [max(1-saturation, 0), 1+saturation]

    def __call__(self, im_lb):
        im = im_lb['im']
        lb = im_lb['lb']
        r_brightness = random.uniform(self.brightness[0], self.brightness[1])
        r_contrast = random.uniform(self.contrast[0], self.contrast[1])
        r_saturation = random.uniform(self.saturation[0], self.saturation[1])
        im = ImageEnhance.Brightness(im).enhance(r_brightness)
        im = ImageEnhance.Contrast(im).enhance(r_contrast)
        im = ImageEnhance.Color(im).enhance(r_saturation)
        return dict(im = im,
                    lb = lb,
                )


class MultiScale(object):
    def __init__(self, scales):
        self.scales = scales

    def __call__(self, img):
        W, H = img.size
        sizes = [(int(W*ratio), int(H*ratio)) for ratio in self.scales]
        imgs = []
        [imgs.append(img.resize(size, Image.BILINEAR)) for size in sizes]
        return imgs


class Compose(object):
    def __init__(self, do_list):
        self.do_list = do_list

    def __call__(self, im_lb):
        for comp in self.do_list:
            im_lb = comp(im_lb)
        return im_lb




if __name__ == '__main__':
    flip = HorizontalFlip(p = 1)
    crop = RandomCrop((321, 321))
    rscales = RandomScale((0.75, 1.0, 1.5, 1.75, 2.0))
    img = Image.open('data/img.jpg')
    lb = Image.open('data/label.png')



import csv
import numpy as np
from pyVHR.datasets.dataset import Dataset
from pyVHR.BPM.BPM import BVPsignal


class UBFC2(Dataset):
    """
    UBFC2 Dataset

    .. UBFC dataset structure:
    .. -----------------
    ..     datasetDIR/
    ..     |   |-- SubjDIR1/
    ..     |       |-- vid.avi
    ..     |...
    ..     |   |-- SubjDIRM/
    ..     |       |-- vid.avi
    """
    name = 'UBFC2'
    signalGT = 'BVP'     # GT signal type
    numLevels = 2        # depth of the filesystem collecting video and BVP files
    numSubjects = 26     # number of subjects
    video_EXT = 'avi'    # extension of the video files
    frameRate = 30       # vieo frame rate
    VIDEO_SUBSTRING = 'vid'  # substring contained in the filename
    SIG_EXT = 'txt'     # extension of the BVP files
    SIG_SUBSTRING = ''  # substring contained in the filename
    SIG_SampleRate = 30  # sample rate of the BVP files
    skinThresh = [40, 60]  # thresholds for skin detection

    def readSigfile(self, filename):
        """ 
        Load signal from file.
        
        Returns:
            a :py:class:`pyVHR.BPM.BPM.BVPsignal` object that can be used to extract BPM signal from ground truth BVP signal.
        """
        gtTrace = []
        gtTime = []
        gtHR = []
        with open(filename, 'r') as f:
            x = f.readlines()

        s = x[0].split(' ')
        s = list(filter(lambda a: a != '', s))
        gtTrace = np.array(s).astype(np.float64)

        t = x[2].split(' ')
        t = list(filter(lambda a: a != '', t))
        gtTime = np.array(t).astype(np.float64)

        hr = x[1].split(' ')
        hr = list(filter(lambda a: a != '', hr))
        gtHR = np.array(hr).astype(np.float64)

        data = np.array(gtTrace)
        time = np.array(gtTime)
        self.SIG_SampleRate = np.round(1/np.mean(np.diff(time)))

        return BVPsignal(data, self.SIG_SampleRate)



import cupy
import numpy as np
from scipy.signal import welch
from scipy.signal import find_peaks, stft
from plotly.subplots import make_subplots
from plotly.colors import n_colors
import plotly.graph_objects as go
from scipy.interpolate import interp1d
from pyVHR.plot.visualize import VisualizeParams
from pyVHR.BPM.utils import *

"""
This module contains all the class and the methods for transforming 
a BVP signal in a BPM signal.
"""


class BVPsignal:
    """
    Manage (multi-channel, row-wise) BVP signals, and transforms it in BPM.
    This class is used to obtain BPM signal from ground truth BVP signal.
    """
    nFFT = 2048  # freq. resolution for STFTs
    step = 1       # step in seconds

    def __init__(self, data, fs, startTime=0, minHz=0.75, maxHz=4., verb=False):
        if len(data.shape) == 1:
            self.data = data.reshape(1, -1)  # 2D array raw-wise
        else:
            self.data = data
        self.fs = fs                       # sample rate
        self.startTime = startTime
        self.verb = verb
        self.minHz = minHz
        self.maxHz = maxHz

    def spectrogram(self, winsize=5):
        """
        Compute the BVP signal spectrogram restricted to the
        band 42-240 BPM by using winsize (in sec) samples.
        """

        # -- spect. Z is 3-dim: Z[#chnls, #freqs, #times]
        F, T, Z = stft(self.data,
                       self.fs,
                       nperseg=self.fs*winsize,
                       noverlap=self.fs*(winsize-self.step),
                       boundary='even',
                       nfft=self.nFFT)
        Z = np.squeeze(Z, axis=0)

        # -- freq subband (0.75 Hz - 4.0 Hz)
        minHz = 0.65
        maxHz = 4.0
        band = np.argwhere((F > minHz) & (F < maxHz)).flatten()
        self.spect = np.abs(Z[band, :])     # spectrum magnitude
        self.freqs = 60*F[band]            # spectrum freq in bpm
        self.times = T                     # spectrum times

        # -- BPM estimate by spectrum
        self.bpm = self.freqs[np.argmax(self.spect, axis=0)]

    def displaySpectrum(self, display=False, dims=3):
        """Show the spectrogram of the BVP signal"""

        # -- check if bpm exists
        try:
            bpm = self.bpm
        except AttributeError:
            self.spectrogram()
            bpm = self.bpm

        t = self.times
        f = self.freqs
        S = self.spect

        fig = go.Figure()
        fig.add_trace(go.Heatmap(z=S, x=t, y=f, colorscale="viridis"))
        fig.add_trace(go.Scatter(
            x=t, y=bpm, name='Frequency Domain', line=dict(color='red', width=2)))

        fig.update_layout(autosize=False, height=420, showlegend=True,
                          title='Spectrogram of the BVP signal',
                          xaxis_title='Time (sec)',
                          yaxis_title='BPM (60*Hz)',
                          legend=dict(
                              x=0,
                              y=1,
                              traceorder="normal",
                              font=dict(
                                family="sans-serif",
                                size=12,
                                color="black"),
                              bgcolor="LightSteelBlue",
                              bordercolor="Black",
                              borderwidth=2)
                          )

        fig.show(renderer=VisualizeParams.renderer)

    def getBPM(self, winsize=5):
        """
        Get the BPM signal extracted from the ground truth BVP signal.
        """
        self.spectrogram(winsize)
        return self.bpm, self.times


class BPMcuda:
    """
    This class transforms a BVP signal in a BPM signal using GPU.

    BVP signal must be a float32 cupy.ndarray with shape [num_estimators, num_frames].
    """
    def __init__(self, data, fps, startTime=0, minHz=0.65, maxHz=4., verb=False):
        """
        Input 'data' is a BVP signal defined as a float32 cupy.ndarray with shape [num_estimators, num_frames]
        """
        self.nFFT = 2048//1  # freq. resolution for STFTs
        if len(data.shape) == 1:
            self.data = data.reshape(1, -1)  # 2D array raw-wise
        else:
            self.data = data
        self.fps = fps                       # sample rate
        self.startTime = startTime
        self.verb = verb
        self.minHz = minHz
        self.maxHz = maxHz
        self.gpuData = cupy.asarray(
            [self.fps, self.nFFT, self.minHz, self.maxHz])
        self.gmodel = Model(gaussian, independent_vars=['x', 'mu', 'a'])

    def BVP_to_BPM(self):
        """
        Return the BPM signal as a float32 cupy.ndarray with shape [num_estimators, ]. Remember that
        the value returned is on GPU, use cupy.asnumpy() to transform it to a Numpy ndarray.

        This method use the Welch's method to estimate the spectral density of the BVP signal,
        then it chooses as BPM the maximum Amplitude frequency.
        """
        # -- interpolation for less than 256 samples
        _, n = self.data.shape
        if self.data.shape[0] == 0:
            return cupy.float32(0.0)
        Pfreqs, Power = Welch_cuda(self.data, self.gpuData[0], self.gpuData[2], self.gpuData[3], self.gpuData[1])
        # -- BPM estimate 
        Pmax = cupy.argmax(Power, axis=1)  # power max
        return Pfreqs[Pmax.squeeze()]

    
    def BVP_to_BPM_PSD_clustering(self, out_fact=1):
        """
        Return the BPM signal as a numpy.float32. 

        TODO: cambiare descrizione.
        This method use the Welch's method to estimate the spectral density of the BVP signal; in case
        of multiple estimators the method sum all the Power Spectums, then it chooses as BPM the 
        maximum Amplitude frequency.
        """
        # -- interpolation for less than 256 samples
        if self.data.shape[0] == 0:
            return np.float32(0.0)
        Pfreqs, Power = Welch_cuda(self.data, self.fps, self.minHz, self.maxHz, self.nFFT)
        # F are frequencies, PSD is Power Spectrum Density of all estimators
        F = cupy.asnumpy(Pfreqs)
        PSD = cupy.asnumpy(Power)
        # Less then 3 estimators -  don't need clustering
        # we choose the maximum Amplitude frequency
        if PSD.shape[0] < 3:
          IDmax = np.unravel_index(np.argmax(PSD, axis=None), PSD.shape)
          Fmax = F[IDmax[1]]
          return np.float32(Fmax)

        W = pairwise_distances(PSD, PSD, metric='cosine')
        theta = circle_clustering(W, eps=0.01)
        # bi-partition, sum and normalization
        PSD0, PSD1, PSD_outlayers, median_PSD0, median_PSD1 = optimize_partition(theta, out_fact=out_fact)

        P0 = np.sum(PSD[PSD0,:], axis=0)
        max = np.max(P0, axis=0)
        max = np.expand_dims(max, axis=0)
        P0 = np.squeeze(np.divide(P0, max))

        P1 = np.sum(PSD[PSD1,:], axis=0)
        max = np.max(P1, axis=0)
        max = np.expand_dims(max, axis=0)
        P1 = np.squeeze(np.divide(P1, max))

        # peaks
        peak0_idx = np.argmax(P0) 
        P0_max = P0[peak0_idx]
        F0 = F[peak0_idx]

        peak1_idx = np.argmax(P1) 
        P1_max = P1[peak1_idx]
        F1 = F[peak1_idx]
    
        # Gaussian fitting
        result0, G1, sigma0 = gaussian_fit(self.gmodel, P0, F, F0, 1)  # Gaussian fit 
        result1, G2, sigma1 = gaussian_fit(self.gmodel, P1, F, F1, 1)  # Gaussian fit 
        chis0 = result0.chisqr
        chis1 = result1.chisqr
        rchis0 = result0.redchi
        rchis1 = result1.redchi
        aic0 = result0.aic
        aic1 = result1.aic
        bic0 = result0.bic
        bic1 = result1.bic

        # ranking
        rankP0 = 0
        rankP1 = 0
        rankP0 = rankP0 + 1 if chis0 < chis1 else rankP0
        rankP1 = rankP1 + 1 if chis0 >= chis1 else rankP1
        rankP0 = rankP0 + 1 if rchis0 < rchis1 else rankP0
        rankP1 = rankP1 + 1 if rchis0 >= rchis1 else rankP1
        rankP0 = rankP0 + 1 if aic0 < aic1 else rankP0
        rankP1 = rankP1 + 1 if aic0 >= aic1 else rankP1
        rankP0 = rankP0 + 1 if bic0 < bic1 else rankP0
        rankP1 = rankP1 + 1 if bic0 >= bic1 else rankP1
        rankP0 = rankP0 + 1 if sigma0 < sigma1 else rankP0
        rankP1 = rankP1 + 1 if sigma0 >= sigma1 else rankP1

        # best fit
        bpm = None
        if rankP0 > rankP1:
            bpm = F0
        else:
            bpm = F1

        return np.float32(bpm)

class BPM:
    """
    This class transforms a BVP signal in a BPM signal using CPU.

    BVP signal must be a float32 numpy.ndarray with shape [num_estimators, num_frames].
    """
    def __init__(self, data, fps, startTime=0, minHz=0.65, maxHz=4., verb=False):
        """
        Input 'data' is a BVP signal defined as a float32 Numpy.ndarray with shape [num_estimators, num_frames]
        """
        self.nFFT = 2048//1  # freq. resolution for STFTs
        if len(data.shape) == 1:
            self.data = data.reshape(1, -1)  # 2D array raw-wise
        else:
            self.data = data
        self.fps = fps                       # sample rate
        self.startTime = startTime
        self.verb = verb
        self.minHz = minHz
        self.maxHz = maxHz
        # PSD_clustering needs a gaussain fitting model
        self.gmodel = Model(gaussian, independent_vars=['x', 'mu', 'a'])

    def BVP_to_BPM(self):
        """
        Return the BPM signal as a float32 Numpy.ndarray with shape [num_estimators, ].

        This method use the Welch's method to estimate the spectral density of the BVP signal,
        then it chooses as BPM the maximum Amplitude frequency.
        """
        if self.data.shape[0] == 0:
            return np.float32(0.0)
        Pfreqs, Power = Welch(self.data, self.fps, self.minHz, self.maxHz, self.nFFT)
        # -- BPM estimate
        Pmax = np.argmax(Power, axis=1)  # power max
        return Pfreqs[Pmax.squeeze()]

    def BVP_to_BPM_PSD_clustering(self, out_fact=1):
        """
        Return the BPM signal as a numpy.float32. 

        TODO: cambiare descrizione.
        This method use the Welch's method to estimate the spectral density of the BVP signal; in case
        of multiple estimators the method sum all the Power Spectums, then it chooses as BPM the 
        maximum Amplitude frequency.
        """
        # -- interpolation for less than 256 samples
        if self.data.shape[0] == 0:
            return np.float32(0.0)
        Pfreqs, Power = Welch(self.data, self.fps, self.minHz, self.maxHz, self.nFFT)
        # F are frequencies, PSD is Power Spectrum Density of all estimators
        F = Pfreqs
        PSD = Power
        # Less then 3 estimators -  don't need clustering
        # we choose the maximum Amplitude frequency
        if PSD.shape[0] < 3:
          IDmax = np.unravel_index(np.argmax(PSD, axis=None), PSD.shape)
          Fmax = F[IDmax[1]]
          return np.float32(Fmax)

        W = pairwise_distances(PSD, PSD, metric='cosine')
        theta = circle_clustering(W, eps=0.01)
        # bi-partition, sum and normalization
        PSD0, PSD1, PSD_outlayers, median_PSD0, median_PSD1 = optimize_partition(theta, out_fact=out_fact)

        P0 = np.sum(PSD[PSD0,:], axis=0)
        max = np.max(P0, axis=0)
        max = np.expand_dims(max, axis=0)
        P0 = np.squeeze(np.divide(P0, max))

        P1 = np.sum(PSD[PSD1,:], axis=0)
        max = np.max(P1, axis=0)
        max = np.expand_dims(max, axis=0)
        P1 = np.squeeze(np.divide(P1, max))

        # peaks
        peak0_idx = np.argmax(P0) 
        P0_max = P0[peak0_idx]
        F0 = F[peak0_idx]

        peak1_idx = np.argmax(P1) 
        P1_max = P1[peak1_idx]
        F1 = F[peak1_idx]
    
        # Gaussian fitting
        result0, G1, sigma0 = gaussian_fit(self.gmodel, P0, F, F0, 1)  # Gaussian fit 
        result1, G2, sigma1 = gaussian_fit(self.gmodel, P1, F, F1, 1)  # Gaussian fit 
        chis0 = result0.chisqr
        chis1 = result1.chisqr
        rchis0 = result0.redchi
        rchis1 = result1.redchi
        aic0 = result0.aic
        aic1 = result1.aic
        bic0 = result0.bic
        bic1 = result1.bic

        # ranking
        rankP0 = 0
        rankP1 = 0
        rankP0 = rankP0 + 1 if chis0 < chis1 else rankP0
        rankP1 = rankP1 + 1 if chis0 >= chis1 else rankP1
        rankP0 = rankP0 + 1 if rchis0 < rchis1 else rankP0
        rankP1 = rankP1 + 1 if rchis0 >= rchis1 else rankP1
        rankP0 = rankP0 + 1 if aic0 < aic1 else rankP0
        rankP1 = rankP1 + 1 if aic0 >= aic1 else rankP1
        rankP0 = rankP0 + 1 if bic0 < bic1 else rankP0
        rankP1 = rankP1 + 1 if bic0 >= bic1 else rankP1
        rankP0 = rankP0 + 1 if sigma0 < sigma1 else rankP0
        rankP1 = rankP1 + 1 if sigma0 >= sigma1 else rankP1

        # best fit
        bpm = None
        if rankP0 > rankP1:
            bpm = F0
        else:
            bpm = F1 

        return np.float32(bpm)


# --------------------------------------------------------------------- #


def multi_est_BPM_median(bpms):
    """
    This method is used for computing the median of a multi-estimators BPM windowed signal.

    Args:
        bpms (list): list of lenght num_windows of BPM signals, each defined as a Numpy.ndarray with shape [num_estimators, ],
            or 1D Numpy.ndarray in case of a single estimator;

    Returns: 
        The median of the multi-estimators BPM signal defined as a float32 Numpy.ndarray with shape [num_windows,]; if a window 
        has num_estimators == 0, then the median value is set to 0.0 .
        The Median Absolute Deviation (MAD) of the multi-estimators BPM signal
    """
    median_bpms = np.zeros((len(bpms),))
    MAD = np.zeros((len(bpms),))
    i = 0
    for bpm in bpms:
        if len(bpm.shape) > 0 and bpm.shape[0] == 0:
            median_bpms[i] = np.float32(0.0)
            MAD[i] = 0.
        else:
            median_bpms[i] = np.float32(np.median(bpm))
            if len(bpm.shape) > 0:
                MAD[i] = np.float32(mad(bpm))                
        i += 1

    return median_bpms, MAD


def BVP_to_BPM(bvps, fps, minHz=0.65, maxHz=4.):
    """
    Transform a BVP windowed signal in a BPM signal using CPU.

    This method use the Welch's method to estimate the spectral density of the BVP signal,
    then it chooses as BPM the maximum Amplitude frequency.

    Args:
        bvps (list): list of length num_windows of BVP signal defined as float32 Numpy.ndarray with shape [num_estimators, num_frames].
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).

    Returns:
        A list of length num_windows of BPM signals defined as a float32 Numpy.ndarray with shape [num_estimators, ].
        If any BPM can't be found in a window, then the ndarray has num_estimators == 0.
        
    """
    bpms = []
    obj = None
    for bvp in bvps:
        if obj is None:
            obj = BPM(bvp, fps, minHz=minHz, maxHz=maxHz)
        else:
            obj.data = bvp
        bpm_es = obj.BVP_to_BPM()
        bpms.append(bpm_es)
    return bpms


def BVP_to_BPM_cuda(bvps, fps, minHz=0.65, maxHz=4.):
    """
    Transform a BVP windowed signal in a BPM signal using GPU.

    This method use the Welch's method to estimate the spectral density of the BVP signal,
    then it chooses as BPM the maximum Amplitude frequency.

    Args:
        bvps (list): list of length num_windows of BVP signal defined as float32 Numpy.ndarray with shape [num_estimators, num_frames].
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).

    Returns:
        A list of length num_windows of BPM signals defined as a float32 Numpy.ndarray with shape [num_estimators, ].
        If any BPM can't be found in a window, then the ndarray has num_estimators == 0.
        
    """
    bpms = []
    obj = None
    for bvp in bvps:
        bvp_device = cupy.asarray(bvp)
        if obj is None:
            obj = BPMcuda(bvp_device, fps, minHz=minHz, maxHz=maxHz)
        else:
            obj.data = bvp_device
        bpm_es = obj.BVP_to_BPM()
        bpm_es = cupy.asnumpy(bpm_es)
        bpms.append(bpm_es)
    return bpms


def BVP_to_BPM_PSD_clustering(bvps, fps, minHz=0.65, maxHz=4., out_fact=1):
    """
    Transform a BVP windowed signal in a BPM signal using CPU.

    TODO: riscrivere descrizione
    This method use the Welch's method to estimate the spectral density of the BVP signal; in case
    of multiple estimators the method sum all the Power Spectums, then it chooses as BPM the 
    maximum Amplitude frequency.

    Args:
        bvps (list): list of length num_windows of BVP signal defined as float32 Numpy.ndarray with shape [num_estimators, num_frames].
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).

    Returns:
        A list of length num_windows of BPM signals defined as a numpy.float32.
        If any BPM can't be found in a window, then the BPM is 0.0.
        
    """
    bpms = []
    obj = None
    for bvp in bvps:
        if obj is None:
            obj = BPM(bvp, fps, minHz=minHz, maxHz=maxHz)
        else:
            obj.data = bvp
        bpm_es = obj.BVP_to_BPM_PSD_clustering(out_fact=out_fact)
        bpms.append(bpm_es)
    return bpms


def BVP_to_BPM_PSD_clustering_cuda(bvps, fps, minHz=0.65, maxHz=4., out_fact=1):
    """
    Transform a BVP windowed signal in a BPM signal using GPU.

    TODO: riscrivere descrizione
    This method use the Welch's method to estimate the spectral density of the BVP signal; in case
    of multiple estimators the method sum all the Power Spectums, then it chooses as BPM the 
    maximum Amplitude frequency.

    Args:
        bvps (list): list of length num_windows of BVP signal defined as float32 Numpy.ndarray with shape [num_estimators, num_frames].
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).

    Returns:
        A list of length num_windows of BPM signals defined as a numpy.float32.
        If any BPM can't be found in a window, then the BPM is 0.0.
        
    """
    bpms = []
    obj = None
    for bvp in bvps:
        if obj is None:
            obj = BPMcuda(bvp, fps, minHz=minHz, maxHz=maxHz)
        else:
            obj.data = bvp
        bpm_es = obj.BVP_to_BPM_PSD_clustering(out_fact=out_fact)
        bpms.append(bpm_es)
    return bpms


from __future__ import print_function
import matplotlib.pyplot as plt
import PIL
from ipywidgets import interact, interactive, fixed, interact_manual
import ipywidgets as widgets
import plotly.graph_objects as go
import plotly.express as px
from PIL import Image
from IPython.display import display, clear_output
import cv2
import mediapipe as mp
import numpy as np
import pyVHR
from pyVHR.extraction.sig_processing import extract_frames_yield
from scipy.signal import welch
import random


from pyVHR.BPM.utils import *


"""
This module defines classes or methods used for plotting outputs.
"""


class VisualizeParams:
    """
    This class contains usefull parameters used by this module.

    The "renderer" variable is used for rendering plots on Jupyter notebook ('notebook')
    or Colab notebook ('colab').
    """
    renderer = 'colab'  # or 'notebook'

def interactive_image_plot(images_list, scaling=1):
    """
    This method create an interactive plot of a list of images. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        images_list (list of ndarray): list of images as ndarray with shape [rows, columns, rgb_channels].
        scaling (float): scale factor useful for enlarging or decreasing the resolution of the image.
    
    """
    if images_list is None or len(images_list) == 0:
        return

    def f(x):
        PIL_image = PIL.Image.fromarray(np.uint8(images_list[x]))
        width, height = PIL_image.size
        PIL_image = PIL_image.resize(
            (int(width*scaling), int(height*scaling)), PIL.Image.NEAREST)
        display(PIL_image)
    interact(f, x=widgets.IntSlider(
        min=0, max=len(images_list)-1, step=1, value=0))


def display_video(video_file_name, scaling=1):
    """
    This method create an interactive plot for visualizing the frames of a video. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        video_file_name (str): video file name or path.
        scaling (float): scale factor useful for enlarging or decreasing the resolution of the image.
    
    """
    original_frames = [cv2.cvtColor(_, cv2.COLOR_BGR2RGB)
                       for _ in extract_frames_yield(video_file_name)]
    interactive_image_plot(original_frames, scaling)


def visualize_windowed_sig(windowed_sig, window):
    """
    This method create a plotly plot for visualizing a window of a windowed signal. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        windowed_sig (list of float32 ndarray): windowed signal as a list of length num_windows of float32 ndarray with shape [num_estimators, rgb_channels, window_frames].
        window (int): the index of the window to plot.
    
    """
    fig = go.Figure()
    i = 1
    for e in windowed_sig[window]:
        name = "sig_" + str(i) + "_r"
        r_color = random.randint(1, 255)
        g_color = random.randint(1, 255)
        b_color = random.randint(1, 255)
        fig.add_trace(go.Scatter(x=np.arange(e.shape[-1]), y=e[0, :],
                                 mode='lines', marker_color='rgba('+str(r_color)+', '+str(g_color)+', '+str(b_color)+', 1.0)', name=name))
        name = "sig_" + str(i) + "_g"
        fig.add_trace(go.Scatter(x=np.arange(e.shape[-1]), y=e[1, :],
                                 mode='lines', marker_color='rgba('+str(r_color)+', '+str(g_color)+', '+str(b_color)+', 1.0)', name=name))
        name = "sig_" + str(i) + "_b"
        fig.add_trace(go.Scatter(x=np.arange(e.shape[-1]), y=e[2, :],
                                 mode='lines', marker_color='rgba('+str(r_color)+', '+str(g_color)+', '+str(b_color)+', 1.0)', name=name))
        i += 1
    fig.update_layout(title="WIN #" + str(window))
    fig.show(renderer=VisualizeParams.renderer)


def visualize_BVPs(BVPs, window):
    """
    This method create a plotly plot for visualizing a window of a windowed BVP signal. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        BVPs (list of float32 ndarray): windowed BPM signal as a list of length num_windows of float32 ndarray with shape [num_estimators, window_frames].
        window (int): the index of the window to plot.
    
    """
    fig = go.Figure()
    i = 1
    bvp = BVPs[window]
    for e in bvp:
        name = "BVP_" + str(i)
        fig.add_trace(go.Scatter(x=np.arange(bvp.shape[1]), y=e[:],
                                 mode='lines', name=name))
        i += 1
    fig.update_layout(title="BVP #" + str(window))
    fig.show(renderer=VisualizeParams.renderer)


def visualize_multi_est_BPM_vs_BPMs_list(multi_est_BPM, BPMs_list):
    """
    This method create a plotly plot for visualizing a multi-estimator BPM signal and a list of BPM signals. 
    This is usefull when comparing Patches BPMs vs Holistic and Ground Truth BPMs. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        multi_est_BPM (list): multi-estimator BPM signal is a list that contains two elements [mul-est-BPM, times]; the first contains BPMs as a list of 
            length num_windows of float32 ndarray with shape [num_estimators, ], the second is a float32 1D ndarray that contains the time in seconds of each BPM.
        BPMs_list (list): The BPM signals is a 2D list structured as [[BPM_list, times, name_tag], ...]. The first element is a float32 ndarray that
            contains the BPM signal, the second element is a float32 1D ndarray that contains the time in seconds of each BPM, the third element is a string
            that is used in the plot's legend.    
    """
    fig = go.Figure()
    for idx, _ in enumerate(BPMs_list):
        name = str(BPMs_list[idx][2])
        fig.add_trace(go.Scatter(x=BPMs_list[idx][1], y=BPMs_list[idx][0],
                                 mode='lines', name=name))
    for w, _ in enumerate(multi_est_BPM[0]):
        name = "multi_est_BPM_" + str(w+1)
        data = multi_est_BPM[0][w]
        if data.shape == ():
            t = [multi_est_BPM[1][w], ]
            data = [multi_est_BPM[0][w], ]
        else:
            t = multi_est_BPM[1][w] * np.ones(data.shape[0])
        fig.add_trace(go.Scatter(x=t, y=data,
                                 mode='markers', marker=dict(size=2,), name=name))
    fig.update_layout(title="BPMs_estimators vs BPMs_list")
    fig.show(renderer=VisualizeParams.renderer)


def visualize_BPMs(BPMs_list):
    """
    This method create a plotly plot for visualizing a list of BPM signals. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        BPMs_list (list): The BPM signals is a 2D list structured as [[BPM_list, times, name_tag], ...]. The first element is a float32 ndarray that
            contains the BPM signal, the second element is a float32 1D ndarray that contains the time in seconds of each BPM, the third element is a string
            that is used in the plot's legend.    
    """
    fig = go.Figure()
    i = 1
    for e in BPMs_list:
        name = str(e[2])
        fig.add_trace(go.Scatter(x=e[1], y=e[0],
                                 mode='lines+markers', name=name))
        i += 1
    fig.update_layout(title="BPMs")
    fig.show(renderer=VisualizeParams.renderer)


def visualize_BVPs_PSD(BVPs, window, fps, minHz=0.65, maxHz=4.0):
    """
    This method create a plotly plot for visualizing the Power Spectral Density of a window of a windowed BVP signal. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        BVPs (list of float32 ndarray): windowed BPM signal as a list of length num_windows of float32 ndarray with shape [num_estimators, window_frames].
        window (int): the index of the window to plot.
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
    
    """
    data = BVPs[window]
    _, n = data.shape
    if data.shape[0] == 0:
        return np.float32(0.0)
    if n < 256:
        seglength = n
        overlap = int(0.8*n)  # fixed overlapping
    else:
        seglength = 256
        overlap = 200
    # -- periodogram by Welch
    F, P = welch(data, nperseg=seglength,
                 noverlap=overlap, fs=fps, nfft=2048)
    F = F.astype(np.float32)
    P = P.astype(np.float32)
    # -- freq subband (0.65 Hz - 4.0 Hz)
    band = np.argwhere((F > minHz) & (F < maxHz)).flatten()
    Pfreqs = 60*F[band]
    Power = P[:, band]
    # -- BPM estimate by PSD
    Pmax = np.argmax(Power, axis=1)  # power max

    # plot
    fig = go.Figure()
    for idx in range(P.shape[0]):
        fig.add_trace(go.Scatter(
            x=F*60, y=P[idx], name="PSD_"+str(idx)+" no band"))
        fig.add_trace(go.Scatter(
            x=Pfreqs, y=Power[idx], name="PSD_"+str(idx)+" band"))
    fig.update_layout(title="PSD #" + str(window))
    fig.show(renderer=VisualizeParams.renderer)


def visualize_landmarks_list(image_file_name=None, landmarks_list=None):
    """
    This method create a plotly plot for visualizing a list of facial landmarks on a given image. This is useful
    for studying and analyzing the available facial points of MediaPipe (https://google.github.io/mediapipe/solutions/face_mesh.html).
    This method must be called inside a Jupyter notebook or a Colab notebook.

    Args:
        image_file_name (str): image file name or path (preferred png or jpg).
        landmarks_list (list): list of positive integers between 0 and 467 that identify patches centers (landmarks).
    
    """
    PRESENCE_THRESHOLD = 0.5
    VISIBILITY_THRESHOLD = 0.5
    if image_file_name is None:
        image_file_name = pyVHR.__path__[0] + '/resources/img/face.png' 
    imag = cv2.imread(image_file_name, cv2.COLOR_RGB2BGR)
    imag = cv2.cvtColor(imag, cv2.COLOR_BGR2RGB)
    mp_drawing = mp.solutions.drawing_utils
    mp_face_mesh = mp.solutions.face_mesh
    with mp_face_mesh.FaceMesh(
            static_image_mode=True,
            max_num_faces=1,
            min_detection_confidence=0.5) as face_mesh:
        image = cv2.imread(image_file_name)
        results = face_mesh.process(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        if not results.multi_face_landmarks:
            return
        width = image.shape[1]
        height = image.shape[0]
        face_landmarks = results.multi_face_landmarks[0]
        ldmks = np.zeros((468, 3), dtype=np.float32)
        for idx, landmark in enumerate(face_landmarks.landmark):
            if ((landmark.HasField('visibility') and landmark.visibility < VISIBILITY_THRESHOLD)
                    or (landmark.HasField('presence') and landmark.presence < PRESENCE_THRESHOLD)):
                ldmks[idx, 0] = -1.0
                ldmks[idx, 1] = -1.0
                ldmks[idx, 2] = -1.0
            else:
                coords = mp_drawing._normalized_to_pixel_coordinates(
                    landmark.x, landmark.y, width, height)
                if coords:
                    ldmks[idx, 0] = coords[0]
                    ldmks[idx, 1] = coords[1]
                    ldmks[idx, 2] = idx
                else:
                    ldmks[idx, 0] = -1.0
                    ldmks[idx, 1] = -1.0
                    ldmks[idx, 2] = -1.0
    
    filtered_ldmks = []
    if landmarks_list is not None:
        for idx in landmarks_list:
            filtered_ldmks.append(ldmks[idx])
        filtered_ldmks = np.array(filtered_ldmks, dtype=np.float32)
    else:
        filtered_ldmks = ldmks

    fig = px.imshow(imag)
    for l in filtered_ldmks:
        name = 'ldmk_' + str(int(l[2]))
        fig.add_trace(go.Scatter(x=(l[0],), y=(l[1],), name=name, mode='markers', 
                                marker=dict(color='blue', size=3)))
    fig.update_xaxes(range=[0,imag.shape[1]])
    fig.update_yaxes(range=[imag.shape[0],0])
    fig.update_layout(paper_bgcolor='#eee') 
    fig.show(renderer=VisualizeParams.renderer)


def visualize_BVPs_PSD_clutering(GT_BPM, GT_times , BVPs, times, fps, minHz=0.65, maxHz=4.0, out_fact=1):
    """
    TODO: documentare
    This method create a plotly plot for visualizing the Power Spectral Density of a window of a windowed BVP signal. This method must be called
    inside a Jupyter notebook or a Colab notebook.

    Args:
        BVPs (list of float32 ndarray): windowed BPM signal as a list of length num_windows of float32 ndarray with shape [num_estimators, window_frames].
        window (int): the index of the window to plot.
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
    
    """
    gmodel = Model(gaussian, independent_vars=['x', 'mu', 'a'])
    for i,X in enumerate(BVPs):
        if X.shape[0] == 0:
            continue
        F, PSD = Welch(X, fps, minHz, maxHz)
        W = pairwise_distances(PSD, PSD, metric='cosine')
        #W = np.corrcoef(PSD)-np.eye(W.shape[0])
        #W = dtw.distance_matrix_fast(PSD.astype(np.double))
        theta = circle_clustering(W, eps=0.01)

        # bi-partition, sum and normalization
        P, Q, Z, med_elem_P, med_elem_Q = optimize_partition(theta, out_fact=out_fact)

        P0 = np.sum(PSD[P,:], axis=0)
        max = np.max(P0, axis=0)
        max = np.expand_dims(max, axis=0)
        P0 = np.squeeze(np.divide(P0, max))

        P1 = np.sum(PSD[Q,:], axis=0)
        max = np.max(P1, axis=0)
        max = np.expand_dims(max, axis=0)
        P1 = np.squeeze(np.divide(P1, max))
        
        # peaks
        peak0_idx = np.argmax(P0) 
        P0_max = P0[peak0_idx]
        F0 = F[peak0_idx]

        peak1_idx = np.argmax(P1) 
        P1_max = P1[peak1_idx]
        F1 = F[peak1_idx]
        
        # Gaussian fitting
        result0, G1, sigma0 = gaussian_fit(gmodel, P0, F, F0, 1)  # Gaussian fit 
        result1, G2, sigma1 = gaussian_fit(gmodel, P1, F, F1, 1)  # Gaussian fit 
        chisqr0 = result0.chisqr
        chisqr1 = result1.chisqr

        print('** Processing window n. ', i)
        t = np.argmin(np.abs(times[i]-GT_times))
        GT = GT_BPM[t]
        print('GT = ', GT, '   freq0 max = ', F0, '   freq1 max = ', F1)
        # TODO: rifare con plotly
        plt.figure(figsize=(14, 4))
        plt.subplot(121)
        plt.plot(F, PSD[P,:].T, alpha=0.2)
        _, top = plt.ylim() 
        plt.plot(F, 0.5*top*P0, color='blue')
        
        plt.subplot(122)
        plt.ylim(0,1.1) 
        plt.plot(F, result0.best_fit, linestyle='-', color='magenta')
        plt.fill_between(F, result0.best_fit, alpha=0.1)
        plt.plot(F, P0, color='blue')
        plt.vlines(F0, ymin=0, ymax=1, linestyle='-.', color='blue')
        plt.vlines(GT,  ymin=0, ymax=1.1, linestyle='-.', color='black')
        plt.title('err = '+ str(np.round(np.abs(GT-F0),2)) +' -- sigma = '+ str(np.round(sigma0,2)) + ' -- chi sqr = ' + str(str(np.round(chisqr0,2))))
        plt.show()

        # second figure
        plt.figure(figsize=(14, 4))
        plt.subplot(121)
        plt.plot(F, PSD[Q,:].T, alpha=0.2)
        _, top = plt.ylim() 
        plt.plot(F, 0.5*top*P1, color='red')
        plt.subplot(122)
        plt.ylim(0,1.1) 
        plt.plot(F, result1.best_fit, linestyle='-', color='magenta')
        plt.fill_between(F, result1.best_fit, alpha=0.1)
        plt.plot(F, P1, color='red')
        plt.vlines(F1, ymin=0, ymax=1, linestyle='-.', color='red')
        plt.vlines(GT,  ymin=0, ymax=1.1, linestyle='-.', color='black')
        plt.title('err = '+ str(np.round(np.abs(GT-F1),2)) +' -- sigma = '+ str(np.round(sigma1,2)) + ' -- chi sqr = ' + str(str(np.round(chisqr1,2))))
        plt.show()

        # centers
        C1 = [np.cos(med_elem_P), np.sin(med_elem_P)]
        C2 = [np.cos(med_elem_Q), np.sin(med_elem_Q)]

        #hist, bins = histogram(theta, nbins=256)
        labels = np.zeros_like(theta)
        labels[Q] = 1
        labels[Z] = 2
        plot_circle(theta, l=labels, C1=C1, C2=C2)


  
def plot_circle(theta, l=None, C1=None, C2=None, radius=500):
    """
    TODO: documentare
    Produce a plot with the locations of all poles and zeros
    """

    x = np.cos(theta)
    y = np.sin(theta)

    fig = go.Figure()
    fig.add_shape(type="circle", xref="x", yref="y", x0=-1, y0=-1, x1=1, y1=1, line=dict(color="black", width=1))
    
    if l is None:
        fig.add_trace(go.Scatter(x=x, y=y,
            mode='markers',
            marker_symbol='circle',
            marker_size=10))
    else:
        ul = np.unique(l)
        cols = ['blue', 'red', 'gray']
        for c,u in zip(cols,ul):
            idx = np.where(u == l)
            fig.add_trace(go.Scatter(x=x[idx], y=y[idx],
                mode='markers',
                marker_symbol='circle',
                marker_color=c, 
                # marker_line_color=cols[c],
                marker_line_width=0, 
                marker_size=10))
        
    # separator plane
    fig.add_trace(go.Scatter(x=[0.0, C1[0]], y=[0.0, C1[1]],
        mode='lines+markers',
        marker_symbol='circle',
        marker_color='blue', 
        marker_line_color='blue',
        marker_line_width=2, 
        marker_size=2))
    fig.add_trace(go.Scatter(x=[0.0, C2[0]], y=[0.0, C2[1]],
        mode='lines+markers',
        marker_symbol='circle',
        marker_color='red', 
        marker_line_color='red',
        marker_line_width=2, 
        marker_size=2))
    
    M = 1.05
    fig.update_xaxes(title='', range=[-M, M])
    fig.update_yaxes(title='', range=[-M, M])
    fig.update_layout(title='clusters', width=radius, height=radius)
    fig.show(renderer=VisualizeParams.renderer)



from numba import prange, njit
import numpy as np
import cv2
import time
import mediapipe as mp
from numba import prange, njit
import os
import numpy as np
import matplotlib.pyplot as plt
import statistics

def get_magic_landmarks():
    """ returns high_priority and mid_priority list of landmarks identification number """
    return [*MagicLandmarks.forehead_center, *MagicLandmarks.cheek_left_bottom, *MagicLandmarks.cheek_right_bottom], [*MagicLandmarks.forehoead_right, *MagicLandmarks.forehead_left, *MagicLandmarks.cheek_left_top, *MagicLandmarks.cheek_right_top]


class MagicLandmarks():
    """
    This class contains usefull lists of landmarks identification numbers.
    """
    high_prio_forehead = [10, 67, 69, 104, 108, 109, 151, 299, 337, 338]
    high_prio_nose = [3, 4, 5, 6, 45, 51, 115, 122, 131, 134, 142, 174, 195, 196, 197, 198,
                      209, 217, 220, 236, 248, 275, 277, 281, 360, 363, 399, 419, 420, 429, 437, 440, 456]
    high_prio_left_cheek = [36, 47, 50, 100, 101, 116, 117,
                            118, 119, 123, 126, 147, 187, 203, 205, 206, 207, 216]
    high_prio_right_cheek = [266, 280, 329, 330, 346, 347,
                             347, 348, 355, 371, 411, 423, 425, 426, 427, 436]

    mid_prio_forehead = [8, 9, 21, 68, 103, 251,
                         284, 297, 298, 301, 332, 333, 372, 383]
    mid_prio_nose = [1, 44, 49, 114, 120, 121, 128, 168, 188, 351, 358, 412]
    mid_prio_left_cheek = [34, 111, 137, 156, 177, 192, 213, 227, 234]
    mid_prio_right_cheek = [340, 345, 352, 361, 454]
    mid_prio_chin = [135, 138, 169, 170, 199, 208, 210, 211,
                     214, 262, 288, 416, 428, 430, 431, 432, 433, 434]
    mid_prio_mouth = [92, 164, 165, 167, 186, 212, 322, 391, 393, 410]
    # more specific areas
    forehead_left = [21, 71, 68, 54, 103, 104, 63, 70,
                     53, 52, 65, 107, 66, 108, 69, 67, 109, 105]
    forehead_center = [10, 151, 9, 8, 107, 336, 285, 55, 8]
    forehoead_right = [338, 337, 336, 296, 285, 295, 282,
                       334, 293, 301, 251, 298, 333, 299, 297, 332, 284]
    eye_right = [283, 300, 368, 353, 264, 372, 454, 340, 448,
                 450, 452, 464, 417, 441, 444, 282, 276, 446, 368]
    eye_left = [127, 234, 34, 139, 70, 53, 124,
                35, 111, 228, 230, 121, 244, 189, 222, 143]
    nose = [193, 417, 168, 188, 6, 412, 197, 174, 399, 456,
            195, 236, 131, 51, 281, 360, 440, 4, 220, 219, 305]
    mounth_up = [186, 92, 167, 393, 322, 410, 287, 39, 269, 61, 164]
    mounth_down = [43, 106, 83, 18, 406, 335, 273, 424, 313, 194, 204]
    chin = [204, 170, 140, 194, 201, 171, 175,
            200, 418, 396, 369, 421, 431, 379, 424]
    cheek_left_bottom = [215, 138, 135, 210, 212, 57, 216, 207, 192]
    cheek_right_bottom = [435, 427, 416, 364,
                          394, 422, 287, 410, 434, 436]
    cheek_left_top = [116, 111, 117, 118, 119, 100, 47, 126, 101, 123,
                      137, 177, 50, 36, 209, 129, 205, 147, 177, 215, 187, 207, 206, 203]
    cheek_right_top = [349, 348, 347, 346, 345, 447, 323,
                       280, 352, 330, 371, 358, 423, 426, 425, 427, 411, 376]
    # dense zones used for convex hull masks
    left_eye = [157,144, 145, 22, 23, 25, 154, 31, 160, 33, 46, 52, 53, 55, 56, 189, 190, 63, 65, 66, 70, 221, 222, 223, 225, 226, 228, 229, 230, 231, 232, 105, 233, 107, 243, 124]
    right_eye = [384, 385, 386, 259, 388, 261, 265, 398, 276, 282, 283, 285, 413, 293, 296, 300, 441, 442, 445, 446, 449, 451, 334, 463, 336, 464, 467, 339, 341, 342, 353, 381, 373, 249, 253, 255]
    mounth = [391, 393, 11, 269, 270, 271, 287, 164, 165, 37, 167, 40, 43, 181, 313, 314, 186, 57, 315, 61, 321, 73, 76, 335, 83, 85, 90, 106]
    # equispaced facial points - mouth and eyes are excluded.
    equispaced_facial_points = [2, 3, 4, 5, 6, 8, 9, 10, 18, 21, 32, 35, 36, 43, 46, 47, 48, 50, 54, \
             58, 67, 68, 69, 71, 92, 93, 101, 103, 104, 108, 109, 116, 117, \
             118, 123, 132, 134, 135, 138, 139, 142, 148, 149, 150, 151, 152, 182, 187, 188, 193, 197, 201, 205, 206, 207, \
             210, 211, 212, 216, 234, 248, 251, 262, 265, 266, 273, 277, 278, 280, \
             284, 288, 297, 299, 322, 323, 330, 332, 333, 337, 338, 345, \
             346, 361, 363, 364, 367, 368, 371, 377, 379, 411, 412, 417, 421, 425, 426, 427, 430, 432, 436]

@njit(parallel=True)
def draw_rects(image, xcenters, ycenters, xsides, ysides, color):
    """
    This method is used to draw N rectangles on a image.
    """
    for idx in prange(len(xcenters)):
        leftx = int(xcenters[idx] - xsides[idx]/2)
        rightx = int(xcenters[idx] + xsides[idx]/2)
        topy = int(ycenters[idx] - ysides[idx]/2)
        bottomy = int(ycenters[idx] + ysides[idx]/2)
        for x in prange(leftx, rightx):
            if topy >= 0 and x >= 0 and x < image.shape[1]:
                image[topy, x, 0] = color[0]
                image[topy, x, 1] = color[1]
                image[topy, x, 2] = color[2]
            if bottomy < image.shape[0] and x >= 0 and x < image.shape[1]:
                image[bottomy, x, 0] = color[0]
                image[bottomy, x, 1] = color[1]
                image[bottomy, x, 2] = color[2]
        for y in prange(topy, bottomy):
            if leftx >= 0 and y >= 0 and y < image.shape[0]:
                image[y, leftx, 0] = color[0]
                image[y, leftx, 1] = color[1]
                image[y, leftx, 2] = color[2]
            if rightx < image.shape[1] and y >= 0 and y < image.shape[0]:
                image[y, rightx, 0] = color[0]
                image[y, rightx, 1] = color[1]
                image[y, rightx, 2] = color[2]
    return image

def sig_windowing(sig, wsize, stride, fps):
    """
    This method is used to divide a RGB signal into overlapping windows.

    Args:
        sig (float32 ndarray): ndarray with shape [num_frames, num_estimators, rgb_channels].
        wsize (float): window size in seconds.
        stride (float): stride between overlapping windows in seconds.
        fps (float): frames per seconds.

    Returns:
        windowed signal as a list of length num_windows of float32 ndarray with shape [num_estimators, rgb_channels, window_frames],
        and a 1D ndarray of times in seconds,where each one is the center of a window.
    """
    N = sig.shape[0]
    block_idx, timesES = sliding_straded_win_offline(N, wsize, stride, fps)
    block_signals = []
    for e in block_idx:
        st_frame = int(e[0])
        end_frame = int(e[-1])
        wind_signal = np.copy(sig[st_frame: end_frame+1])
        wind_signal = np.swapaxes(wind_signal, 0, 1)
        wind_signal = np.swapaxes(wind_signal, 1, 2)
        block_signals.append(wind_signal)
    return block_signals, timesES

def raw_windowing(raw_signal, wsize, stride, fps):
    """
    This method is used to divide a Raw signal into overlapping windows.

    Args:
        sig (float32 ndarray): ndarray of images with shape [num_frames, rows, columns, rgb_channels].
        wsize (float): window size in seconds.
        stride (float): stride between overlapping windows in seconds.
        fps (float): frames per seconds.

    Returns:
        windowed signal as a list of length num_windows of float32 ndarray with shape [num_frames, rows, columns, rgb_channels],
        and a 1D ndarray of times in seconds,where each one is the center of a window.
    """
    N = raw_signal.shape[0]
    block_idx, timesES = sliding_straded_win_offline(N, wsize, stride, fps)
    block_signals = []
    for e in block_idx:
        st_frame = int(e[0])
        end_frame = int(e[-1])
        wind_signal = np.copy(raw_signal[st_frame: end_frame+1])
        block_signals.append(wind_signal)
    return block_signals, timesES

def sliding_straded_win_offline(N, wsize, stride, fps):
    """
    This method is used to compute all the info for creating an overlapping windows signal.

    Args:
        N (int): length of the signal.
        wsize (float): window size in seconds.
        stride (float): stride between overlapping windows in seconds.
        fps (float): frames per seconds.

    Returns:
        List of ranges, each one contains the indices of a window, and a 1D ndarray of times in seconds, where each one is the center of a window.
    """
    wsize_fr = wsize*fps
    stride_fr = stride*fps
    idx = []
    timesES = []
    num_win = int((N-wsize_fr)/stride_fr)+1
    s = 0
    for i in range(num_win):
        idx.append(np.arange(s, s+wsize_fr))
        s += stride_fr
        timesES.append(wsize/2+stride*i)
    return idx, np.array(timesES, dtype=np.float32)

def get_fps(videoFileName):
    """
    This method returns the fps of a video file name or path.
    """
    vidcap = cv2.VideoCapture(videoFileName)
    fps = vidcap.get(cv2.CAP_PROP_FPS)
    vidcap.release()
    return fps

def extract_frames_yield(videoFileName):
    """
    This method yield the frames of a video file name or path.
    """
    vidcap = cv2.VideoCapture(videoFileName)
    success, image = vidcap.read()
    while success:
        yield image
        success, image = vidcap.read()
    vidcap.release()


import cupy
import math
import time
import numpy as np
import torch
import os
from sklearn.decomposition import PCA
from pyVHR.BVP.utils import jadeR


"""
This module contains a collection of known rPPG methods.

rPPG METHOD SIGNATURE
An rPPG method must accept theese parameters:
    > signal -> RGB signal as float32 ndarray with shape [num_estimators, rgb_channels, num_frames], or a custom signal.
    > **kargs [OPTIONAL] -> usefull parameters passed to the filter method.
It must return a BVP signal as float32 ndarray with shape [num_estimators, num_frames].
"""


# ------------------------------------------------------------------------------------- #
#                                     rPPG METHODS                                      #
# ------------------------------------------------------------------------------------- #


def cpu_CHROM(signal):
    """
    CHROM method on CPU using Numpy.

    De Haan, G., & Jeanne, V. (2013). Robust pulse rate from chrominance-based rPPG. IEEE Transactions on Biomedical Engineering, 60(10), 2878-2886.
    """
    X = signal
    Xcomp = 3*X[:, 0] - 2*X[:, 1]
    Ycomp = (1.5*X[:, 0])+X[:, 1]-(1.5*X[:, 2])
    sX = np.std(Xcomp, axis=1)
    sY = np.std(Ycomp, axis=1)
    alpha = (sX/sY).reshape(-1, 1)
    alpha = np.repeat(alpha, Xcomp.shape[1], 1)
    bvp = Xcomp - np.multiply(alpha, Ycomp)
    return bvp


def cupy_CHROM(signal):
    """
    CHROM method on GPU using Cupy.

    De Haan, G., & Jeanne, V. (2013). Robust pulse rate from chrominance-based rPPG. IEEE Transactions on Biomedical Engineering, 60(10), 2878-2886.
    """
    X = signal
    Xcomp = 3*X[:, 0] - 2*X[:, 1]
    Ycomp = (1.5*X[:, 0])+X[:, 1]-(1.5*X[:, 2])
    sX = cupy.std(Xcomp, axis=1)
    sY = cupy.std(Ycomp, axis=1)
    alpha = (sX/sY).reshape(-1, 1)
    alpha = cupy.repeat(alpha, Xcomp.shape[1], 1)
    bvp = Xcomp - cupy.multiply(alpha, Ycomp)
    return bvp


def torch_CHROM(signal):
    """
    CHROM method on CPU using Torch.

    De Haan, G., & Jeanne, V. (2013). Robust pulse rate from chrominance-based rPPG. IEEE Transactions on Biomedical Engineering, 60(10), 2878-2886.
    """
    X = signal
    Xcomp = 3*X[:, 0] - 2*X[:, 1]
    Ycomp = (1.5*X[:, 0])+X[:, 1]-(1.5*X[:, 2])
    sX = torch.std(Xcomp, axis=1)
    sY = torch.std(Ycomp, axis=1)
    alpha = (sX/sY).reshape(-1, 1)
    alpha = torch.repeat_interleave(alpha, Xcomp.shape[1], 1)
    bvp = Xcomp - torch.mul(alpha, Ycomp)
    return bvp


def cpu_LGI(signal):
    """
    LGI method on CPU using Numpy.

    Pilz, C. S., Zaunseder, S., Krajewski, J., & Blazek, V. (2018). Local group invariance for heart rate estimation from face videos in the wild. In Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition Workshops (pp. 1254-1262).
    """
    X = signal
    U, _, _ = np.linalg.svd(X)
    S = U[:, :, 0]
    S = np.expand_dims(S, 2)
    sst = np.matmul(S, np.swapaxes(S, 1, 2))
    p = np.tile(np.identity(3), (S.shape[0], 1, 1))
    P = p - sst
    Y = np.matmul(P, X)
    bvp = Y[:, 1, :]
    return bvp


def cpu_POS(signal, **kargs):
    """
    POS method on CPU using Numpy.

    The dictionary parameters are: {'fps':float}.

    Wang, W., den Brinker, A. C., Stuijk, S., & de Haan, G. (2016). Algorithmic principles of remote PPG. IEEE Transactions on Biomedical Engineering, 64(7), 1479-1491. 
    """
    # Run the pos algorithm on the RGB color signal c with sliding window length wlen
    # Recommended value for wlen is 32 for a 20 fps camera (1.6 s)
    eps = 10**-9
    X = signal
    e, c, f = X.shape            # e = #estimators, c = 3 rgb ch., f = #frames
    w = int(1.6 * kargs['fps'])   # window length

    # stack e times fixed mat P
    P = np.array([[0, 1, -1], [-2, 1, 1]])
    Q = np.stack([P for _ in range(e)], axis=0)

    # Initialize (1)
    H = np.zeros((e, f))
    for n in np.arange(w, f):
        # Start index of sliding window (4)
        m = n - w + 1
        # Temporal normalization (5)
        Cn = X[:, :, m:(n + 1)]
        M = 1.0 / (np.mean(Cn, axis=2)+eps)
        M = np.expand_dims(M, axis=2)  # shape [e, c, w]
        Cn = np.multiply(M, Cn)

        # Projection (6)
        S = np.dot(Q, Cn)
        S = S[0, :, :, :]
        S = np.swapaxes(S, 0, 1)    # remove 3-th dim

        # Tuning (7)
        S1 = S[:, 0, :]
        S2 = S[:, 1, :]
        alpha = np.std(S1, axis=1) / (eps + +np.std(S2, axis=1))
        alpha = np.expand_dims(alpha, axis=1)
        Hn = np.add(S1, alpha * S2)
        Hnm = Hn - np.expand_dims(np.mean(Hn, axis=1), axis=1)
        # Overlap-adding (8)
        H[:, m:(n + 1)] = np.add(H[:, m:(n + 1)], Hnm)

    return H


def cupy_POS(signal, **kargs):
    """
    POS method on GPU using Cupy.

    The dictionary parameters are: {'fps':float}.

    Wang, W., den Brinker, A. C., Stuijk, S., & de Haan, G. (2016). Algorithmic principles of remote PPG. IEEE Transactions on Biomedical Engineering, 64(7), 1479-1491. 
    """
    # Run the pos algorithm on the RGB color signal c with sliding window length wlen
    # Recommended value for wlen is 32 for a 20 fps camera (1.6 s)
    X = signal
    fps = cupy.float32(kargs['fps'])
    e, c, f = X.shape            # e = #estimators, c = 3 rgb ch., f = #frames
    w = int(1.6 * fps)   # window length

    # stack e times fixed mat P
    P = cupy.array([[0, 1, -1], [-2, 1, 1]])
    Q = cupy.stack([P for _ in range(e)], axis=0)

    # Initialize (1)
    H = cupy.zeros((e, f))
    for n in cupy.arange(w, f):
        # Start index of sliding window (4)
        m = n - w + 1
        # Temporal normalization (5)
        Cn = X[:, :, m:(n + 1)]
        M = 1.0 / (cupy.mean(Cn, axis=2))
        M = cupy.expand_dims(M, axis=2)  # shape [e, c, w]
        Cn = cupy.multiply(M, Cn)

        # Projection (6)
        S = cupy.dot(Q, Cn)
        S = S[0, :, :, :]
        S = cupy.swapaxes(S, 0, 1)    # remove 3-th dim

        # Tuning (7)
        S1 = S[:, 0, :]
        S2 = S[:, 1, :]
        alpha = cupy.std(S1, axis=1) / (cupy.std(S2, axis=1))
        alpha = cupy.expand_dims(alpha, axis=1)
        Hn = cupy.add(S1, alpha * S2)
        Hnm = Hn - cupy.expand_dims(cupy.mean(Hn, axis=1), axis=1)
        # Overlap-adding (8)
        H[:, m:(n + 1)] = cupy.add(H[:, m:(n + 1)], Hnm)

    return H


def cpu_PBV(signal):
    """
    PBV method on CPU using Numpy.

    De Haan, G., & Van Leest, A. (2014). Improved motion robustness of remote-PPG by using the blood volume pulse signature. Physiological measurement, 35(9), 1913.
    """
    sig_mean = np.mean(signal, axis = 2)

    signal_norm_r = signal[:,0,:] / np.expand_dims(sig_mean[:,0],axis=1)
    signal_norm_g = signal[:,1,:] / np.expand_dims(sig_mean[:,1],axis=1)
    signal_norm_b = signal[:,2,:] / np.expand_dims(sig_mean[:,2],axis=1)

    pbv_n = np.array([np.std(signal_norm_r, axis = 1), np.std(signal_norm_g, axis = 1), np.std(signal_norm_b, axis = 1)])
    pbv_d = np.sqrt(np.var(signal_norm_r, axis = 1) + np.var(signal_norm_g, axis = 1) + np.var(signal_norm_b, axis = 1))
    pbv = pbv_n / pbv_d

    C = np.swapaxes(np.array([signal_norm_r, signal_norm_g, signal_norm_b]),0,1)
    Ct =np.swapaxes(np.swapaxes(np.transpose(C),0,2),1,2)
    Q = np.matmul(C, Ct)
    W = np.linalg.solve(Q,np.swapaxes(pbv,0,1))

    A = np.matmul(Ct, np.expand_dims(W,axis = 2))
    B =  np.matmul(np.swapaxes(np.expand_dims(pbv.T,axis=2),1,2),np.expand_dims(W,axis = 2))
    bvp = A / B
    return bvp.squeeze(axis=2)


def cpu_PCA(signal,**kargs):
    """
    PCA method on CPU using Numpy.

    The dictionary parameters are {'component':str}. Where 'component' can be 'second_comp' or 'all_comp'.

    Lewandowska, M., Rumiński, J., Kocejko, T., & Nowak, J. (2011, September). Measuring pulse rate with a webcam—a non-contact method for evaluating cardiac activity. In 2011 federated conference on computer science and information systems (FedCSIS) (pp. 405-410). IEEE.
    """
    bvp = []
    for i in range(signal.shape[0]):
        X = signal[i]
        pca = PCA(n_components=3)
        pca.fit(X)

        # selector
        if kargs['component']=='all_comp':
            bvp.append(pca.components_[0] * pca.explained_variance_[0])
            bvp.append(pca.components_[1] * pca.explained_variance_[1])
        elif kargs['component']=='second_comp':
            bvp.append(pca.components_[1] * pca.explained_variance_[1])
    bvp = np.array(bvp)
    return bvp

def cpu_GREEN(signal):
    """
    GREEN method on CPU using Numpy

    Verkruysse, W., Svaasand, L. O., & Nelson, J. S. (2008). Remote plethysmographic imaging using ambient light. Optics express, 16(26), 21434-21445.
    """
    return signal[:,1,:]

def cpu_ICA(signal, **kargs):
    """
    ICA method on CPU using Numpy.

    The dictionary parameters are {'component':str}. Where 'component' can be 'second_comp' or 'all_comp'.

    Poh, M. Z., McDuff, D. J., & Picard, R. W. (2010). Non-contact, automated cardiac pulse measurements using video imaging and blind source separation. Optics express, 18(10), 10762-10774.    
    """
    bvp = []
    for X in signal:
        W = jadeR(X, verbose=False)  
        bvp.append(np.dot(W,X))
    
    # selector
    bvp = np.array(bvp)
    l, c, f = bvp.shape     # l=#landmks c=#3chs, f=#frames
    if kargs['component']=='all_comp':
        bvp = np.reshape(bvp, (l*c, f))  # compact into 2D matrix 
    elif kargs['component']=='second_comp':
        bvp = np.reshape(bvp[:,1,:], (l, f))
    
    # collect
    return bvp

def cpu_SSR(raw_signal,**kargs):
    """
    SSR method on CPU using Numpy.

    'raw_signal' is a float32 ndarray with shape [num_frames, rows, columns, rgb_channels]; it can be obtained by
    using the :py:class:‵pyVHR.extraction.sig_processing.SignalProcessing‵ class ('extract_raw_holistic' method).

    The dictionary parameters are: {'fps':float}.

    Wang, W., Stuijk, S., & De Haan, G. (2015). A novel algorithm for remote photoplethysmography: Spatial subspace rotation. IEEE transactions on biomedical engineering, 63(9), 1974-1984.
    """
    # utils functions #
    def __build_p(τ, k, l, U, Λ):
        """
        builds P
        Parameters
        ----------
        k: int
            The frame index
        l: int
            The temporal stride to use
        U: numpy.ndarray
            The eigenvectors of the c matrix (for all frames up to counter).
        Λ: numpy.ndarray
            The eigenvalues of the c matrix (for all frames up to counter).
        Returns
        -------
        p: numpy.ndarray
            The p signal to add to the pulse.
        """
        # SR'
        SR = np.zeros((3, l), np.float32)  # dim: 3xl
        z = 0

        for t in range(τ, k, 1):  # 6, 7
            a = Λ[0, t]
            b = Λ[1, τ]
            c = Λ[2, τ]
            d = U[:, 0, t].T
            e = U[:, 1, τ]
            f = U[:, 2, τ]
            g = U[:, 1, τ].T
            h = U[:, 2, τ].T
            x1 = a / b
            x2 = a / c
            x3 = np.outer(e, g)
            x4 = np.dot(d, x3)
            x5 = np.outer(f, h)
            x6 = np.dot(d, x5)
            x7 = np.sqrt(x1)
            x8 = np.sqrt(x2)
            x9 = x7 * x4
            x10 = x8 * x6
            x11 = x9 + x10
            SR[:, z] = x11  # 8 | dim: 3
            z += 1

        # build p and add it to the final pulse signal
        s0 = SR[0, :]  # dim: l
        s1 = SR[1, :]  # dim: l
        p = s0 - ((np.std(s0) / np.std(s1)) * s1)  # 10 | dim: l
        p = p - np.mean(p)  # 11
        return p  # dim: l
        
    def __build_correlation_matrix(V):
        # V dim: (W×H)x3
        #V = np.unique(V, axis=0)
        V_T = V.T  # dim: 3x(W×H)
        N = V.shape[0]
        # build the correlation matrix
        C = np.dot(V_T, V)  # dim: 3x3
        C = C / N

        return C

    def __eigs(C):
        """
        get eigenvalues and eigenvectors, sort them.
        Parameters
        ----------
        C: numpy.ndarray
            The RGB values of skin-colored pixels.
        Returns
        -------
        Λ: numpy.ndarray
            The eigenvalues of the correlation matrix
        U: numpy.ndarray
            The (sorted) eigenvectors of the correlation matrix
        """
        # get eigenvectors and sort them according to eigenvalues (largest first)
        L, U = np.linalg.eig(C)  # dim Λ: 3 | dim U: 3x3
        idx = L.argsort()  # dim: 3x1
        idx = idx[::-1]  # dim: 1x3
        L_ = L[idx]  # dim: 3
        U_ = U[:, idx]  # dim: 3x3

        return L_, U_
    # ----------------------------------- #

    fps = int(kargs['fps'])

    raw_sig = raw_signal
    K = len(raw_sig)
    l = int(fps)

    P = np.zeros(K)  # 1 | dim: K
    # store the eigenvalues Λ and the eigenvectors U at each frame
    L = np.zeros((3, K), dtype=np.float32)  # dim: 3xK
    U = np.zeros((3, 3, K), dtype=np.float32)  # dim: 3x3xK

    for k in range(K):
        n_roi = len(raw_sig[k])
        VV = []
        V = raw_sig[k].astype(np.float32)
        idx = V!=0
        idx2 = np.logical_and(np.logical_and(idx[:,:,0], idx[:,:,1]), idx[:,:,2])
        V_skin_only = V[idx2]
        VV.append(V_skin_only)
        
        VV = np.vstack(VV)

        C = __build_correlation_matrix(VV)  #dim: 3x3

        # get: eigenvalues Λ, eigenvectors U
        L[:,k], U[:,:,k] = __eigs(C)  # dim Λ: 3 | dim U: 3x3

        # build p and add it to the pulse signal P
        if k >= l:  # 5
            tau = k - l  # 5
            p = __build_p(tau, k, l, U, L)  # 6, 7, 8, 9, 10, 11 | dim: l
            P[tau:k] += p  # 11

        if np.isnan(np.sum(P)):
            print('NAN')
            print(raw_sig[k])
            
    bvp = P
    bvp = np.expand_dims(bvp,axis=0)
    return bvp


from numba import cuda, njit, prange
import cupy
import math
import numpy as np
import torchvision.transforms as transforms
import torch
from numba import prange, njit, cuda
from pyVHR.resources.faceparsing.model import BiSeNet
import os
import pyVHR
from scipy.spatial import ConvexHull
from PIL import Image, ImageDraw
import requests

"""
This module defines classes or methods used for skin extraction.
"""

### functions and parameters ###

class SkinProcessingParams():
    """
        This class contains usefull parameters used by this module.

        RGB_LOW_TH (numpy.int32): RGB low-threshold value.

        RGB_HIGH_TH (numpy.int32): RGB high-threshold value.
    """
    RGB_LOW_TH = np.int32(55)
    RGB_HIGH_TH = np.int32(200)


def bbox2_CPU(img):
    """
    Args:
        img (ndarray): ndarray with shape [rows, columns, rgb_channels].

    Returns: 
        Four cropping coordinates (row, row, column, column) for removing black borders (RGB [O,O,O]) from img.
    """
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    nzrows = np.nonzero(rows)
    nzcols = np.nonzero(cols)
    if nzrows[0].size == 0 or nzcols[0].size == 0:
        return -1, -1, -1, -1
    rmin, rmax = np.nonzero(rows)[0][[0, -1]]
    cmin, cmax = np.nonzero(cols)[0][[0, -1]]
    return rmin, rmax, cmin, cmax


def bbox2_GPU(img):
    """
    Args:
        img (cupy.ndarray): cupy.ndarray with shape [rows, columns, rgb_channels].

    Returns: 
        Four cropping coordinates (row, row, column, column) for removing black borders (RGB [O,O,O]) from img. 
        The returned variables are on GPU.
    """
    rows = cupy.any(img, axis=1)
    cols = cupy.any(img, axis=0)
    nzrows = cupy.nonzero(rows)
    nzcols = cupy.nonzero(cols)
    if nzrows[0].size == 0 or nzcols[0].size == 0:
        return -1, -1, -1, -1
    rmin, rmax = cupy.nonzero(rows)[0][[0, -1]]
    cmin, cmax = cupy.nonzero(cols)[0][[0, -1]]
    return rmin, rmax, cmin, cmax


### SKIN EXTRACTION CLASSES ###

class SkinExtractionFaceParsing():
    """
        This class performs skin extraction on CPU/GPU using Face Parsing.
        https://github.com/zllrunning/face-parsing.PyTorch
    """
    def __init__(self, device='CPU'):
        """
        Args:
            device (str): This class can execute code on 'CPU' or 'GPU'.
        """
        self.device = device
        n_classes = 19
        self.net = BiSeNet(n_classes=n_classes)
        if self.device == 'GPU':
            self.net.cuda()
            self.kernel_cuda_skin_copy_and_filter = kernel_cuda_skin_copy_and_filter()
        save_pth = os.path.dirname(
            pyVHR.resources.faceparsing.model.__file__) + "/79999_iter.pth"
        if not os.path.isfile(save_pth):
            url = "https://github.com/phuselab/pyVHR/blob/main/pyVHR/resources/faceparsing/79999_iter.pth"
            print('Downloading faceparsing model...')
            r = requests.get(url, allow_redirects=True)
            open(save_pth, 'wb').write(r.content)    
        self.net.load_state_dict(torch.load(save_pth))
        self.net.eval()
        self.to_tensor = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        ])

    def extract_skin(self, image, ldmks):
        """
        This method extract the skin from an image using Face Parsing. 
        Landmarks (ldmks) are used to create a facial bounding box for cropping the face; this way
        the network used in Face Parsing is more accurate. 

        Args:
            image (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
            ldmks (float32 ndarray): ndarray with shape [num_landmarks, xy_coordinates].

        Returns:
            Cropped skin-image and non-cropped skin-image; both are uint8 ndarray with shape [rows, columns, rgb_channels].
        """
        # crop with bounding box of ldmks; the network works better if the bounding box is bigger
        aviable_ldmks = ldmks[ldmks[:,0] >= 0][:,:2]  
        min_y, min_x = np.min(aviable_ldmks, axis=0)
        max_y, max_x = np.max(aviable_ldmks, axis=0)
        min_y *= 0.90
        min_x *= 0.90
        max_y = max_y * 1.10 if max_y * 1.10 < image.shape[0] else image.shape[0]
        max_x = max_x * 1.10 if max_x * 1.10 < image.shape[1] else image.shape[1]
        cropped_image = np.copy(image[int(min_y):int(max_y),int(min_x):int(max_x) ,:])
        nda_im = np.array(cropped_image)
        # prepare the image for the bisenet network
        cropped_image = self.to_tensor(cropped_image)
        cropped_image = torch.unsqueeze(cropped_image, 0)
        cropped_skin_img = self.extraction(cropped_image, nda_im)
        # recreate full image using cropped_skin_img
        full_skin_image = np.zeros_like(image)
        full_skin_image[int(min_y):int(max_y),int(min_x):int(max_x) ,:] = cropped_skin_img
        return cropped_skin_img, full_skin_image

    def extraction(self, im, nda_im):
        """
        This method performs skin extraction using Face Parsing.

        Args:
            im (torch.Tensor): torch.Tensor with size [rows, columns, rgb_channels] 
            nda_im (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].

        Returns:
            skin-image as uint8 ndarray with shape [rows, columns, rgb_channels].
        """
        with torch.no_grad():
            if self.device == 'CPU':
                ### bisenet skin detection ###
                out = self.net(im)[0]
                ### gpu cuda skin copy ###
                parsing = out.squeeze(0).argmax(0).numpy()
                parsing = parsing.astype(np.int32)
                nda_im = nda_im.astype(np.uint8)
                # kernel skin copy
                return kernel_skin_copy_and_filter(nda_im, parsing, np.int32(SkinProcessingParams.RGB_LOW_TH), np.int32(SkinProcessingParams.RGB_HIGH_TH))
            else:
                ### bisenet skin detection ###
                im = im.cuda()
                out = self.net(im)[0]
                ### gpu cuda skin copy ###
                dev_parsing = out.squeeze(0).argmax(0).type(torch.int32)
                dev_nda_im = cuda.to_device(nda_im)
                dev_new_im = cuda.to_device(np.zeros_like(nda_im))
                low_high_f = cuda.to_device(
                    np.array([SkinProcessingParams.RGB_LOW_TH, SkinProcessingParams.RGB_HIGH_TH], dtype=np.int32))
                # define number of blocks and threads per block
                threadsperblock = (16, 16)
                blockspergrid_x = math.ceil(nda_im.shape[0] / threadsperblock[0])
                blockspergrid_y = math.ceil(nda_im.shape[1] / threadsperblock[1])
                blockspergrid = (blockspergrid_x, blockspergrid_y)
                # kernel invoke
                self.kernel_cuda_skin_copy_and_filter[blockspergrid, threadsperblock](
                    dev_nda_im, dev_parsing, dev_new_im, low_high_f)
                # copy to CPU result
                newimg = dev_new_im.copy_to_host()
                # free memory
                im = None
                out = None
                dev_nda_im = None
                dev_parsing = None
                dev_new_im = None
                return newimg

# SkinExtractionFaceParsing class kernels #
def kernel_cuda_skin_copy_and_filter():
    """
    Return a Numba cuda.jit kernel defined as:

    @cuda.jit('void(uint8[:,:,:], int32[:,:], uint8[:,:,:], int32[:])')
    def __kernel_cuda_skin_copy_and_filter(orig, pars, new, low_high_filter):
        ''
        This method removes pixels from the image 'orig' that are not skin, or 
        that are outside the RGB range [low_high_filter[0], low_high_filter[1]] (extremes are included).
        ''
  
    This method is important for users who do not use a GPU, beacause they can't compile @cuda.jit.
    """

    @cuda.jit('void(uint8[:,:,:], int32[:,:], uint8[:,:,:], int32[:])')
    def __kernel_cuda_skin_copy_and_filter(orig, pars, new, low_high_filter):
        """
        This method removes pixels from the image 'orig' that are not skin, or 
        that are outside the RGB range [low_high_filter[0], low_high_filter[1]] (extremes are included).
        """
        x, y = cuda.grid(2)
        if x < orig.shape[0] and y < orig.shape[1]:
            # skin class = 1, nose = 10
            if pars[x, y] == 1 or pars[x, y] == 10:
                if not ((orig[x, y, 0] <= low_high_filter[0] and orig[x, y, 1] <= low_high_filter[0] and orig[x, y, 2] <= low_high_filter[0]) or
                        (orig[x, y, 0] >= low_high_filter[1] and orig[x, y, 1] >= low_high_filter[1] and orig[x, y, 2] >= low_high_filter[1])):
                    new[x, y, 0] = orig[x, y, 0]
                    new[x, y, 1] = orig[x, y, 1]
                    new[x, y, 2] = orig[x, y, 2]

    return __kernel_cuda_skin_copy_and_filter


@njit('uint8[:,:,:](uint8[:,:,:], int32[:,:], int32, int32)', parallel=True, nogil=True)
def kernel_skin_copy_and_filter(orig, pars, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method removes pixels from the image 'orig' that are not skin, or 
    that are outside the RGB range [RGB_LOW_TH, RGB_HIGH_TH] (extremes are included).
    """
    new = np.zeros_like(orig)
    for x in prange(orig.shape[0]):
        for y in prange(orig.shape[1]):
            # skin class = 1, nose = 10
            if pars[x, y] == 1 or pars[x, y] == 10:
                if not ((orig[x, y, 0] <= RGB_LOW_TH and orig[x, y, 1] <= RGB_LOW_TH and orig[x, y, 2] <= RGB_LOW_TH) or
                        (orig[x, y, 0] >= RGB_HIGH_TH and orig[x, y, 1] >= RGB_HIGH_TH and orig[x, y, 2] >= RGB_HIGH_TH)):
                    new[x, y, 0] = orig[x, y, 0]
                    new[x, y, 1] = orig[x, y, 1]
                    new[x, y, 2] = orig[x, y, 2]
    return new

class SkinExtractionConvexHull:
    """
        This class performs skin extraction on CPU/GPU using a Convex Hull segmentation obtained from facial landmarks.
    """
    def __init__(self,device='CPU'):
        """
        Args:
            device (str): This class can execute code on 'CPU' or 'GPU'.
        """
        self.device = device
    
    def extract_skin(self,image, ldmks):
        """
        This method extract the skin from an image using Convex Hull segmentation.

        Args:
            image (uint8 ndarray): ndarray with shape [rows, columns, rgb_channels].
            ldmks (float32 ndarray): landmarks used to create the Convex Hull; ldmks is a ndarray with shape [num_landmarks, xy_coordinates].

        Returns:
            Cropped skin-image and non-cropped skin-image; both are uint8 ndarray with shape [rows, columns, rgb_channels].
        """
        from pyVHR.extraction.sig_processing import MagicLandmarks
        aviable_ldmks = ldmks[ldmks[:,0] >= 0][:,:2]        
        # face_mask convex hull 
        hull = ConvexHull(aviable_ldmks)
        verts = [(aviable_ldmks[v,0], aviable_ldmks[v,1]) for v in hull.vertices]
        img = Image.new('L', image.shape[:2], 0)
        ImageDraw.Draw(img).polygon(verts, outline=1, fill=1)
        mask = np.array(img)
        mask = np.expand_dims(mask,axis=0).T

        # left eye convex hull
        left_eye_ldmks = ldmks[MagicLandmarks.left_eye]
        aviable_ldmks = left_eye_ldmks[left_eye_ldmks[:,0] >= 0][:,:2]
        if len(aviable_ldmks) > 3:
            hull = ConvexHull(aviable_ldmks)
            verts = [(aviable_ldmks[v,0], aviable_ldmks[v,1]) for v in hull.vertices]
            img = Image.new('L', image.shape[:2], 0)
            ImageDraw.Draw(img).polygon(verts, outline=1, fill=1)
            left_eye_mask = np.array(img)
            left_eye_mask = np.expand_dims(left_eye_mask,axis=0).T
        else:
            left_eye_mask = np.ones((image.shape[0], image.shape[1],1),dtype=np.uint8)

        # right eye convex hull
        right_eye_ldmks = ldmks[MagicLandmarks.right_eye]
        aviable_ldmks = right_eye_ldmks[right_eye_ldmks[:,0] >= 0][:,:2]
        if len(aviable_ldmks) > 3:
            hull = ConvexHull(aviable_ldmks)
            verts = [(aviable_ldmks[v,0], aviable_ldmks[v,1]) for v in hull.vertices]
            img = Image.new('L', image.shape[:2], 0)
            ImageDraw.Draw(img).polygon(verts, outline=1, fill=1)
            right_eye_mask = np.array(img)
            right_eye_mask = np.expand_dims(right_eye_mask,axis=0).T
        else:
            right_eye_mask = np.ones((image.shape[0], image.shape[1],1),dtype=np.uint8)

        # mounth convex hull
        mounth_ldmks = ldmks[MagicLandmarks.mounth]
        aviable_ldmks = mounth_ldmks[mounth_ldmks[:,0] >= 0][:,:2]
        if len(aviable_ldmks) > 3:
            hull = ConvexHull(aviable_ldmks)
            verts = [(aviable_ldmks[v,0], aviable_ldmks[v,1]) for v in hull.vertices]
            img = Image.new('L', image.shape[:2], 0)
            ImageDraw.Draw(img).polygon(verts, outline=1, fill=1)
            mounth_mask = np.array(img)
            mounth_mask = np.expand_dims(mounth_mask,axis=0).T
        else:
            mounth_mask = np.ones((image.shape[0], image.shape[1],1),dtype=np.uint8)

        # apply masks and crop 
        if self.device == 'GPU':
            image = cupy.asarray(image)
            mask = cupy.asarray(mask)
            left_eye_mask = cupy.asarray(left_eye_mask)
            right_eye_mask = cupy.asarray(right_eye_mask)
            mounth_mask = cupy.asarray(mounth_mask)
        skin_image = image * mask * (1-left_eye_mask) * (1-right_eye_mask) * (1-mounth_mask)

        if self.device == 'GPU':
            rmin, rmax, cmin, cmax = bbox2_GPU(skin_image)
        else:
            rmin, rmax, cmin, cmax = bbox2_CPU(skin_image)

        cropped_skin_im = skin_image
        if rmin >= 0 and rmax >= 0 and cmin >= 0 and cmax >= 0 and rmax-rmin >= 0 and cmax-cmin >= 0:
            cropped_skin_im = skin_image[int(rmin):int(rmax), int(cmin):int(cmax)]

        if self.device == 'GPU':
            cropped_skin_im = cupy.asnumpy(cropped_skin_im)
            skin_image = cupy.asnumpy(skin_image)

        return cropped_skin_im, skin_image
